"""
Handles deploy logic for components
"""
import copy
import importlib
import json
import logging
import os
import sys
import threading
import yaml
import concurrent.futures

from .config import merge_cfgs
from .images import get_is_configs, import_images
from .utils import apply_template, load_cfg_file, trigger_builds, wait_for_ready_threaded
from .secrets import import_secrets, SecretImporter
from .templates import Template, get_templates_in_dir


log = logging.getLogger(__name__)


def deploy_components(
    project_name,
    template_dir,
    components,
    variables_per_component,
    wait=True,
    timeout=300,
    resources_scale_factor=1.0,
    label=None,
):
    """
    Deploy a collection of components in a service-set to a project

    This will:
        1. deploy templates from {template_dir} which match each component name in {components}
            these templates will be processed with {variables}
        2. if {wait} is true, waits for any newly configured DeploymentConfig's to go 'active'

    Args:
        project_name (str) -- name of openshift project
        template_dir (str) -- path to template directory
        components (tuple of tuples) -- list of component names
            component name should match the template filename (minus the .yml/.json)
        variables_per_component (dict) -- multiple key/val pairs with:
            key: component name
            val: dict of variables to pass into 'oc process' as parameters
        images (dict) -- key/val pairs of
            key: name of image stream (as it appears in 'oc get is')
            val: full name of image to import (e.g. the full 'docker pull' name)
        secrets (list of str) -- names of secrets this service set needs to have present
        wait (boolean) -- wait for all deploymentConfigurations/builds to be 'ready'
        timeout (int) -- timeout to wait for all deploymentConfigurations to be 'ready'
        resources_scale_factor (float) -- factor to scale cpu/memory resource requests/limits by
        label (str) -- Label to apply to all deployed resources

    Returns:
        The templates we used to deploy this service set, in the same dict format as
        get_templates_in_dir()
    """
    resources_to_wait_for = []
    templates_by_name = get_templates_in_dir(template_dir)
    processed_templates_by_name = {}

    for comp_name in components:
        if comp_name not in templates_by_name:
            raise ValueError(
                "Component '{}' not found in template dir '{}'".format(comp_name, template_dir)
            )

        template = templates_by_name.get(comp_name)
        template.process(variables_per_component.get(comp_name, {}), resources_scale_factor, label)

        if not template.processed_content:
            log.info("Component %s has an empty template, skipping...", comp_name)
            continue

        processed_templates_by_name[comp_name] = template

        log.info("Deploying component '%s'", comp_name)
        apply_template(project_name, template)

        # Mark certain resources in this component as ones we need to wait on
        for restype in ("deployment", "deploymentconfig", "statefulset", "daemonset"):
            resources_to_wait_for.extend(
                [(restype, name) for name in template.get_processed_names_for_restype(restype)]
            )

        # Re-trigger any builds for deployed build configs
        bcs = template.get_processed_items_for_restype("bc")
        if bcs:
            resources_to_wait_for.extend(trigger_builds(bcs))

    # Wait on all resources that have been marked as 'resources to wait for'
    if wait:
        wait_for_ready_threaded(resources_to_wait_for, timeout=timeout, exit_on_err=True)

    return processed_templates_by_name


def _deploy_dry_run(
    project_name,
    template_dir,
    components,
    variables_per_component,
    label,
    resources_scale_factor,
    jinja_only,
):
    """
    Similar to deploy_components, but only processes the templates.

    Does not actually push any config
    """
    templates_by_name = get_templates_in_dir(template_dir)
    processed_templates_by_name = {}

    for comp_name in components:
        if comp_name not in templates_by_name:
            raise ValueError(
                "Component '{}' not found in template dir '{}'".format(comp_name, template_dir)
            )

        template = templates_by_name.get(comp_name)
        if jinja_only:
            template.process_jinja(variables_per_component.get(comp_name, {}))

        else:
            template.process(
                variables_per_component.get(comp_name, {}), resources_scale_factor, label,
            )

        content_attr_name = "processed_content"
        if jinja_only:
            content_attr_name = "processed_jinja_content"
        if not getattr(template, content_attr_name):
            log.info("Component %s has an empty template, skipping...", comp_name)
            continue

        processed_templates_by_name[comp_name] = template

    return processed_templates_by_name


def deploy_dry_run(
    project_name,
    template_dir,
    components,
    variables_per_component,
    wait=False,
    timeout=None,
    resources_scale_factor=1.0,
    label=None,
):
    return _deploy_dry_run(
        project_name,
        template_dir,
        components,
        variables_per_component,
        label,
        resources_scale_factor,
        jinja_only=False,
    )


def deploy_dry_run_jinja_only(
    project_name,
    template_dir,
    components,
    variables_per_component,
    wait=False,
    timeout=None,
    resources_scale_factor=1.0,
    label=None,
):
    return _deploy_dry_run(
        project_name,
        template_dir,
        components,
        variables_per_component,
        label,
        resources_scale_factor,
        jinja_only=True,
    )


DEFAULT_DEPLOY_METHODS = (None, deploy_components, None)


def _load_module(path, service_set):
    importlib.invalidate_caches()
    spec = importlib.util.spec_from_file_location(f"deploy_{service_set}", path)

    if not spec or not os.path.exists(path):
        return None

    module = importlib.util.module_from_spec(spec)
    sys.modules[f"deploy_{service_set}"] = module
    spec.loader.exec_module(module)
    log.info("Custom script found in '%s' for service set '%s'", path, service_set)

    return module


def _get_custom_deploy_methods(service_set, service_set_dir, root_custom_dir):
    """
    Look for custom deploy module and import its methods.
    """
    module = _load_module(os.path.join(service_set_dir, "custom", "deploy.py"), service_set)
    if not module:
        module = _load_module(
            os.path.join(root_custom_dir, f"deploy_{service_set}.py"), service_set
        )

    if not module:
        module = _load_module(os.path.join(root_custom_dir, "deploy.py"), service_set)

    if not module:
        log.exception("Error loading custom deploy script, using default deploy methods")
        return DEFAULT_DEPLOY_METHODS

    pre_deploy_method = deploy_method = post_deploy_method = None

    try:
        pre_deploy_method = getattr(module, "pre_deploy")
        log.info("Custom pre_deploy() found for service set '%s'", service_set)
    except AttributeError:
        pre_deploy_method = None

    try:
        deploy_method = getattr(module, "deploy")
        log.info("Custom deploy() method found for service set '%s'", service_set)
    except AttributeError:
        deploy_method = deploy_components

    try:
        post_deploy_method = getattr(module, "post_deploy")
        log.info("Custom post_deploy() method found for service set '%s'", service_set)
    except AttributeError:
        post_deploy_method = None

    log.info(
        "Service set '%s' custom pre_deploy(): %s, custom deploy(): %s, custom post_deploy(): %s",
        service_set,
        bool(pre_deploy_method),
        bool(deploy_method),
        bool(post_deploy_method),
    )

    return pre_deploy_method, deploy_method, post_deploy_method


def _get_deploy_methods(config, service_set_name, service_set_dir, root_custom_dir):
    if config.get("custom_deploy_logic", False):
        pre_deploy_method, deploy_method, post_deploy_method = _get_custom_deploy_methods(
            service_set_name, service_set_dir, root_custom_dir
        )
    else:
        pre_deploy_method, deploy_method, post_deploy_method = DEFAULT_DEPLOY_METHODS
    return pre_deploy_method, deploy_method, post_deploy_method


def generate_dry_run_content(all_processed_templates, output="yaml", to_dir=None, jinja_only=False):
    """
    Write processed template content to output directory, or print to stdout if no dir given.
    """
    content_attr_name = "processed_content"
    if jinja_only:
        content_attr_name = "processed_jinja_content"

    if to_dir:
        to_dir = os.path.abspath(to_dir)
        try:
            os.makedirs(to_dir, exist_ok=True)
            log.info("Writing processed templates to output directory: %s", to_dir)
        except OSError as exc:
            log.error("Unable to create output directory '%s': %s", to_dir, str(exc))
            return

    for service_set, processed_templates in all_processed_templates.items():
        for template_name, template_obj in processed_templates.items():
            # Check if the template_obj is an actual template... the "_imagestreams" component
            # added during dry run is just a plain dict
            if isinstance(template_obj, Template):
                content = getattr(template_obj, content_attr_name)
            else:
                content = template_obj

            if not content:
                log.warning("Template '%s' had no processed content", template_name)
            else:
                if output not in ["yaml", "json"]:
                    output = "yaml"
                if output == "yaml":
                    text = yaml.dump(content, default_flow_style=False)
                else:
                    text = json.dumps(content, indent=2)

                if to_dir:
                    service_set_dir = os.path.join(to_dir, service_set)
                    os.makedirs(service_set_dir, exist_ok=True)
                    file_path = os.path.join(service_set_dir, "{}.{}".format(template_name, output))
                    with open(file_path, "w") as f:
                        f.write(text)
                else:
                    print("---\n# {}/{}".format(service_set, template_name))
                    print(text)


class DeployRunner(object):
    def __init__(
        self,
        template_dir,
        project_name,
        env_config_handler,
        ignore_requires,
        service_sets_selected,
        resources_scale_factor,
        root_custom_dir,
        specific_components=None,
        label=None,
        skip=None,
        dry_run=False,
        dry_run_opts=None,
        concurrent=False,
        threadpool_size=os.cpu_count(),
    ):
        self.template_dir = template_dir
        self.root_custom_dir = root_custom_dir
        self.project_name = project_name
        self.ignore_requires = ignore_requires
        self.service_sets_selected = service_sets_selected
        self.resources_scale_factor = resources_scale_factor
        self._deployed_service_sets = []
        self.specific_components = specific_components or []
        self._pick_service_sets = [comp.split("/")[0] for comp in self.specific_components]
        self.label = label
        self.skip = skip
        self.dry_run = dry_run
        self.dry_run_opts = dry_run_opts or {}
        self.env_config_handler = env_config_handler
        self._base_is_configs = {}
        self.concurrent = concurrent
        self.threadpool_size = threadpool_size

    def _get_variables(self, service_set_name, service_set_dir, component):
        variables = {}
        if self.env_config_handler:
            variables = self.env_config_handler.get_vars_for_component(
                service_set_dir, service_set_name, component
            )

        # ocdeployer adds the "NAMESPACE" and "SECRETS_PROJECT" parameter by default at deploy time
        if "parameters" not in variables:
            variables["parameters"] = {}
        variables["parameters"].update(
            {"NAMESPACE": self.project_name, "SECRETS_PROJECT": SecretImporter.source_project}
        )

        return variables

    def _get_variables_per_component(self, set_cfg, service_set_dir, service_set_name):
        variables_per_component = {}
        for _, stage_config in set_cfg.get("deploy_order", {}).items():
            variables_per_component.update(
                {
                    component_name: self._get_variables(
                        service_set_name, service_set_dir, component_name
                    )
                    for component_name in stage_config.get("components", [])
                }
            )
        return variables_per_component

    def _check_requires(self, set_cfg, service_set_name):
        requires = set_cfg.get("requires", [])
        for required_set in requires:
            if required_set not in self._deployed_service_sets:
                raise ValueError(
                    "Config for '{}' requires set '{}' which has not been deployed yet.".format(
                        service_set_name, required_set
                    )
                )

    def _enter_stage(
        self,
        deploy_func,
        components,
        variables_per_component,
        stage,
        deploy_order,
        service_set,
        dir_path,
    ):
        log.info("Entering stage '%s' of config in service set '%s'", stage, service_set)

        # If a component has been skipped, remove it from our component list
        if self.skip:
            for entry in self.skip:
                entry_service_set, entry_component = entry.split("/")
                if entry_service_set == service_set and entry_component in components:
                    log.info("SKIPPING deploy for component: %s", entry_component)
                    components.remove(entry_component)

        # Make sure all the component names have a template
        templates_found = get_templates_in_dir(dir_path)
        for comp in components:
            if comp not in templates_found:
                raise ValueError(
                    "File for component named '{}' does not exist in service set '{}'".format(
                        comp, service_set
                    )
                )

        log.info("Running deploy() in stage '%s' of service set '%s'", stage, service_set)

        processed_templates_this_stage = deploy_func(
            project_name=self.project_name,
            template_dir=dir_path,
            components=components,
            variables_per_component=variables_per_component,
            wait=deploy_order[stage].get("wait", True) is True,
            timeout=deploy_order[stage].get("timeout", 300),
            resources_scale_factor=self.resources_scale_factor,
            label=self.label,
        )
        return processed_templates_this_stage

    def _deploy_stage(
        self, deploy_func, variables_per_component, stage, deploy_order, service_set, dir_path
    ):
        components = deploy_order[stage].get("components", [])
        comps_with_set_name = [f"{service_set}/{comp}" for comp in components]

        if service_set in self.service_sets_selected:
            components_to_deploy = components
        elif self.specific_components:
            # If a component has been 'picked', deploy it
            components_to_deploy = [
                c.split("/")[1] for c in self.specific_components if c in comps_with_set_name
            ]

        if not components_to_deploy:
            log.info(
                "Skipping stage '%s', no selected components are part of this stage", stage,
            )
            return {}

        return self._enter_stage(
            deploy_func,
            components_to_deploy,
            variables_per_component,
            stage,
            deploy_order,
            service_set,
            dir_path,
        )

    def _get_base_cfg(self):
        cfg_path = os.path.join(self.template_dir, "_cfg.yml")
        if not os.path.isdir(self.template_dir) or not os.path.isfile(cfg_path):
            raise ValueError(f"Unable to find config at {cfg_path}")

        # Load _cfg.yml from templates dir
        base_cfg = load_cfg_file(cfg_path)
        # Check if base env files have '_cfg' defined
        base_env_cfg = {}
        if self.env_config_handler:
            base_env_cfg = self.env_config_handler.get_base_env_cfg()
        # Merge the values from the two _cfg definitions, with env settings taking precedence
        return merge_cfgs(base_cfg, base_env_cfg)

    def _get_service_set_cfg(self, service_set, dir_path):
        cfg_path = os.path.join(dir_path, "_cfg.yml")

        if not os.path.isdir(dir_path) or not os.path.isfile(cfg_path):
            raise ValueError(f"Unable to find config for service set at {cfg_path}")

        # Load _cfg.yml from service set
        set_cfg = load_cfg_file(cfg_path)
        # Check if service set env files have '_cfg' defined
        set_env_cfg = {}
        if self.env_config_handler:
            set_env_cfg = self.env_config_handler.get_service_set_env_cfg(dir_path, service_set)
        # Merge the values from the two _cfg definitions, with env settings taking precedence
        return merge_cfgs(set_cfg, set_env_cfg)

    def _deploy_service_set(self, service_set):
        if self.concurrent:
            threading.current_thread().name = service_set[:21]

        log.info("Handling config for service set '%s'", service_set)

        processed_templates = {}

        dir_path = os.path.join(self.template_dir, service_set)
        set_cfg = self._get_service_set_cfg(service_set, dir_path)

        if not self.ignore_requires:
            self._check_requires(set_cfg, service_set)

        if self.dry_run:
            log.info("Doing a DRY RUN of deployment")
            jinja_only = self.dry_run_opts.get("jinja_only")
            deploy_dry_run_func = deploy_dry_run_jinja_only if jinja_only else deploy_dry_run
            pre_deploy_func, deploy_func, post_deploy_func = None, deploy_dry_run_func, None

            is_configs = copy.deepcopy(self._base_is_configs)
            new_is_configs = get_is_configs(set_cfg, self.env_config_handler.env_names)
            for istag, is_config in new_is_configs.items():
                # If an istag is imported by the base _cfg, due to the way ImageImporter works,
                # it will not be re-imported again. So don't overwrite that istag in our
                # processed output either.
                if istag not in is_configs:
                    is_configs[istag] = is_config

            if is_configs:
                processed_templates["_imagestreams"] = {
                    "apiVersion": "v1",
                    "kind": "List",
                    "items": [config for _, config in is_configs.items()],
                }

        else:
            import_secrets(set_cfg, self.env_config_handler.env_names)
            import_images(set_cfg, self.env_config_handler.env_names)

            pre_deploy_func, deploy_func, post_deploy_func = _get_deploy_methods(
                set_cfg, service_set, dir_path, self.root_custom_dir
            )

        variables_per_component = self._get_variables_per_component(set_cfg, dir_path, service_set)

        deploy_order = set_cfg.get("deploy_order", {})

        if pre_deploy_func:
            log.info("Running pre_deploy() for service set '%s'", service_set)
            pre_deploy_func(
                project_name=self.project_name,
                template_dir=dir_path,
                variables_per_component=variables_per_component,
            )

        for stage in sorted(deploy_order.keys()):
            processed_templates.update(
                self._deploy_stage(
                    deploy_func, variables_per_component, stage, deploy_order, service_set, dir_path
                )
            )

        if post_deploy_func:
            log.info("Running post_deploy() for service set '%s'", service_set)
            post_deploy_func(
                processed_templates=processed_templates,
                project_name=self.project_name,
                template_dir=dir_path,
                variables_per_component=variables_per_component,
                timeout=int(set_cfg.get("post_deploy_timeout", 0)),
            )

        self._deployed_service_sets.append(service_set)
        return processed_templates

    def _deploy_sets_concurrently(self, service_sets_selected):
        all_processed_templates = {}
        failed_sets = []

        log.info(
            "deploying service sets (%s) in this stage concurrently, threadpool size: %d",
            ", ".join(service_sets_selected),
            self.threadpool_size,
        )

        with concurrent.futures.ThreadPoolExecutor(max_workers=self.threadpool_size) as executor:
            futures = {
                executor.submit(self._deploy_service_set, service_set): service_set
                for service_set in service_sets_selected
            }
            for future in concurrent.futures.as_completed(futures):
                service_set = futures[future]
                try:
                    all_processed_templates[service_set] = future.result()
                except Exception:
                    log.exception("Service set '%s' hit exception", service_set)
                    failed_sets.append(service_set)

        if failed_sets:
            log.error("deploys failed for service set(s): %s", ", ".join(failed_sets))
            raise Exception("deploys failed in stage")

        return all_processed_templates

    def run(self):
        """
        Load the "_cfg.yml" at {dir_path}, parse it, and deploy the defined components
        according to the defined stages.

        The base template dir can contain folders of other components (called service sets)
        which each have their own _cfg.yml

        Components from each stage are deployed and we wait for them to all reach
        a "ready" state before moving on to the next stage.

        If 'return_immediately' is set to 'True' under a stage, then we will not wait.
        """
        self._deployed_service_sets = []

        base_cfg = self._get_base_cfg()
        deploy_order = base_cfg.get("deploy_order", {})

        self._base_is_configs = get_is_configs(base_cfg, self.env_config_handler.env_names)
        if not self.dry_run:
            import_secrets(base_cfg, self.env_config_handler.env_names)
            import_images(base_cfg, self.env_config_handler.env_names)

        # Verify all service sets exist
        all_service_sets = []
        for stage, stage_data in deploy_order.items():
            all_service_sets.extend(stage_data.get("components", []))

        sets_for_deploy = set(self.service_sets_selected + self._pick_service_sets)

        for service_set in sets_for_deploy:
            if service_set not in all_service_sets:
                raise ValueError("Service set '{}' not found in base config.".format(service_set))

        # Deploy the service sets in proper order
        all_processed_templates = {}

        for stage in sorted(deploy_order.keys()):
            service_sets = deploy_order[stage].get("components", [])
            service_sets_selected = []
            for service_set in service_sets:
                if service_set not in sets_for_deploy:
                    log.info(
                        "Skipping service set '%s', not selected for deploy at runtime", service_set
                    )
                    continue
                service_sets_selected.append(service_set)

            if self.concurrent:
                all_processed_templates.update(
                    self._deploy_sets_concurrently(service_sets_selected)
                )
            else:
                for service_set in service_sets_selected:
                    all_processed_templates[service_set] = self._deploy_service_set(service_set)

        total_components = sum([len(templates) for _, templates in all_processed_templates.items()])
        log.info("Processed templates for %d total components", total_components)
        if not total_components:
            raise ValueError("The chosen combination of '-s' or '-p' did not match any components")

        if self.dry_run:
            generate_dry_run_content(all_processed_templates, **self.dry_run_opts)



#!/usr/bin/env python
from __future__ import print_function

import click
import logging
import os
import json
import sys
import re
import threading

import prompter
import yaml

from ocdeployer.utils import (
    all_sets,
    oc,
    get_dir,
    get_routes,
    switch_to_project,
    get_server_info,
)
from ocdeployer.secrets import SecretImporter
from ocdeployer.deploy import DeployRunner
from ocdeployer.env import EnvConfigHandler, LegacyEnvConfigHandler
from ocdeployer.events import start_event_watcher


log = logging.getLogger("ocdeployer")
CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


def wipe(no_confirm, project, label):
    server = get_server_info()
    extra_msg = ""
    if label:
        extra_msg = " with label '{}'".format(label)

    if not no_confirm and prompter.yesno(
        "I'm about to delete everything in project '{}'{} on server {} -- continue?".format(
            project, extra_msg, server
        ),
        default="no",
    ):
        sys.exit(0)

    switch_to_project(project)

    if label:
        args = ["--selector", label]
    else:
        args = ["--all"]

    oc("delete", "all", *args, _exit_on_err=False)
    oc("delete", "configmap", *args, _exit_on_err=False)
    oc("delete", "secret", *args, _exit_on_err=False)
    oc("delete", "pvc", *args, _exit_on_err=False)


def list_routes(project, output=None):
    switch_to_project(project)
    route_data = get_routes()

    if not output:
        log.info("The following routes now exist:")
        max_len = 10
        for svc_name in route_data:
            if len(svc_name) > max_len:
                max_len = len(svc_name)
        for svc_name, svc_route in route_data.items():
            log.info("%-*s %s", max_len, svc_name, svc_route)

    elif output == "json":
        print(json.dumps(route_data, indent=2))

    elif output == "yaml":
        print(yaml.dump(route_data, default_flow_style=False))


def list_sets(template_dir, output=None):
    as_dict = {"service_sets": all_sets(template_dir)}

    if not output:
        log.info("Available service sets:\n * %s", "\n * ".join(as_dict["service_sets"]))

    elif output == "json":
        print(json.dumps(as_dict, indent=2))

    elif output == "yaml":
        print(yaml.dump(as_dict, default_flow_style=False))


def verify_label(label):
    if not label:
        return
    if not re.match(r"^\w+=\w+$", label):
        log.error("Label '%s' is not valid.  Example: 'mylabel=myvalue'", label)
        sys.exit(1)


@click.group(
    help="Deploys components to a given cluster. NOTE: You need the openshift cli tool"
    " ('oc') installed and to login to your openshift cluster before running the tool.",
    context_settings=CONTEXT_SETTINGS,
)
def main():
    """Main ocdeployer group"""
    threading.current_thread().name = "main"
    logging.basicConfig(
        level=logging.INFO,
        datefmt="%Y-%m-%d %H:%M:%S",
        format="%(asctime)s %(levelname)8s [%(threadName)-15.15s] %(message)s",
    )
    logging.getLogger("sh").setLevel(logging.CRITICAL)


# Options shared by both the "deploy" command and the "process" command
_common_options = [
    click.option("--all", "-a", "all_services", is_flag=True, help="Deploy all service sets"),
    click.option(
        "--sets", "-s", help="Comma,separated,list of specific service set names to deploy"
    ),
    click.option(
        "--pick",
        "-p",
        default=None,
        help=(
            "Comma,separated,list of specific service_set/component to deploy."
            "  E.g. '-p myset/myvm'"
        ),
    ),
    click.option("--skip", "-k", help="Comma,separated,list of service_set/component to skip"),
    click.option(
        "--env",
        "-e",
        "env_values",
        help=(
            "Name of environment to load variables from (default: None).  Use this option multiple"
            " times to concatenate environment configurations. The env listed first takes priority."
            "  You can also specify filenames here (see 'Environment Files' in README)."
        ),
        multiple=True,
    ),
    click.option(
        "--env-file",
        "env_files",
        help=("(legacy) for backward compatibility. Same as using '--env' with a filename."),
        multiple=True,
    ),
    click.option(
        "--template-dir", "-t", default=None, help="Template directory (default 'templates')"
    ),
    click.option(
        "--scale-resources",
        type=float,
        default=1.0,
        help="Factor to scale configured cpu/memory resource requests/limits by",
    ),
    click.option(
        "--concurrent",
        "-c",
        is_flag=True,
        help="Deploy service sets within a stage concurrently using thread pool",
    ),
    click.option(
        "--threadpool-size",
        "-z",
        type=int,
        default=os.cpu_count(),
        help="Threadpool size when running concurrent deploys (default: os.cpu_count()",
    ),
]


def common_options(func):
    """Click decorator used for common options, shared by deploy and process commands."""
    for option in reversed(_common_options):
        func = option(func)
    return func


def output_option(func):
    """Click decorator used for output option, shared by several commands."""
    option = click.option(
        "--output",
        "-o",
        default=None,
        type=click.Choice(["yaml", "json"]),
        help="Output data using yaml or json format",
    )
    return option(func)


def _parse_args(template_dir, env_values, env_files, all_services, sets, pick, dst_project):
    """Parses args common to 'process' and 'deploy'."""
    template_dir = get_dir(template_dir, "templates", "template")

    # Analyze the values provided by --env to determine which config handler we are using
    if env_values and env_files:
        log.error("You cannot use both --env and --env-file")
        sys.exit(1)
    elif env_values:
        all_env_values_are_files = all([os.path.exists(value) for value in env_values])
        some_env_values_are_files = any([os.path.exists(value) for value in env_values])
        if all_env_values_are_files:
            log.info("A specific filename was provided for env, using legacy env file processing")
            env_config_handler = LegacyEnvConfigHandler(env_files=env_values)
        elif some_env_values_are_files:
            log.error("Error: Values for '--env' must be either all filenames, or all env names")
            sys.exit(1)
        else:
            env_config_handler = EnvConfigHandler(env_names=env_values)
    elif env_files:
        log.info("A specific filename was provided for env, using legacy env file processing")
        env_config_handler = LegacyEnvConfigHandler(env_files=env_files)
    else:
        env_config_handler = EnvConfigHandler(env_names=[])

    log.info("Using environments: %s", ", ".join(env_values or env_files))

    if not all_services and not sets and not pick:
        log.error(
            "Error: no service sets or components selected for deploy."
            " Use --sets, --all, or --pick"
        )
        sys.exit(1)

    server = get_server_info()

    sets_selected = []
    specific_components = []

    if all_services:
        sets_selected = all_sets(template_dir)
    else:
        if pick:
            try:
                [p.split("/")[1] for p in pick.split(",")]
            except (ValueError, IndexError):
                log.error("Invalid format for '--pick', use: 'service_set/component'")
                sys.exit(1)
            specific_components = list(set(pick.split(",")))
        if sets:
            sets_selected = list(set(sets.split(",")))

    joined_sets = ", ".join(sets_selected)
    joined_comps = ", ".join(specific_components)
    sets_string = f" service sets: [{joined_sets}]" if joined_sets else ""
    comps_string = f" components: [{joined_comps}]" if joined_comps else ""
    confirm_msg = (
        f"Deploying{sets_string}{comps_string} "
        f"to project '{dst_project}' on server {server} -- continue?"
    )

    return template_dir, env_config_handler, specific_components, sets_selected, confirm_msg


@main.command("process", help="Process templates but do not deploy")
@common_options
@output_option
@click.option(
    "--to-dir",
    default=None,
    help="Save processed templates to specific output directory (default: print to stdout)",
)
@click.option(
    "-j",
    "--jinja-only",
    default=False,
    is_flag=True,
    help="Only provide the output of processing jinja, excludes 'oc process'",
)
@click.argument("dst_project")
def deploy_dry_run(
    dst_project,
    sets,
    all_services,
    env_values,
    env_files,
    template_dir,
    scale_resources,
    pick,
    skip,
    output,
    to_dir,
    jinja_only,
    concurrent,
    threadpool_size,
):
    template_dir, env_config_handler, specific_components, sets_selected, _ = _parse_args(
        template_dir, env_values, env_files, all_services, sets, pick, dst_project
    )

    # No need to set up SecretImporter, it won't be used in a dry run

    DeployRunner(
        template_dir,
        dst_project,
        env_config_handler,
        ignore_requires=True,  # ignore for a dry run
        service_sets_selected=sets_selected,
        resources_scale_factor=scale_resources,
        root_custom_dir=None,  # won't be used in a dry run
        specific_components=specific_components,
        label=None,
        skip=skip.split(",") if skip else None,
        dry_run=True,
        dry_run_opts={"output": output, "to_dir": to_dir, "jinja_only": jinja_only},
        concurrent=concurrent,
        threadpool_size=threadpool_size,
    ).run()


@main.command("deploy", help="Deploy to project")
@common_options
@click.option("--no-confirm", "-f", is_flag=True, help="Do not prompt for confirmation")
@click.option(
    "--secrets-local-dir", default=None, help="Import secrets from local files in a directory",
)
@click.option(
    "--secrets-src-project", default=None, help="Openshift project to import secrets from",
)
@click.option(
    "--ignore-requires",
    "-i",
    is_flag=True,
    help="Ignore the 'requires' statement in config files and deploy anyway",
)
@click.option(
    "--custom-dir",
    "-u",
    "root_custom_dir",
    default=None,
    help="(legacy) specify root custom deploy scripts directory (default 'custom')",
)
@click.option(
    "--label",
    "-l",
    default=None,
    help="Adds a label to each deployed resource.  E.g. '-l app=test'",
)
@click.option(
    "--watch", "-w", is_flag=True, default=False, help="Enable event watching during the deploy"
)
@click.argument("dst_project")
def deploy_to_project(
    dst_project,
    no_confirm,
    secrets_local_dir,
    sets,
    all_services,
    secrets_src_project,
    env_values,
    env_files,
    template_dir,
    ignore_requires,
    scale_resources,
    root_custom_dir,
    pick,
    label,
    skip,
    watch,
    concurrent,
    threadpool_size,
):
    root_custom_dir = get_dir(root_custom_dir, "custom", "custom scripts", optional=True)

    if not dst_project:
        log.error("Error: no destination project given")
        sys.exit(1)

    verify_label(label)

    if secrets_local_dir:
        SecretImporter.local_dir = get_dir(secrets_local_dir, "secrets", "secrets", optional=True)
    if secrets_src_project:
        SecretImporter.source_project = secrets_src_project

    template_dir, env_config_handler, specific_components, sets_selected, confirm_msg = _parse_args(
        template_dir, env_values, env_files, all_services, sets, pick, dst_project
    )

    if not no_confirm and not prompter.yesno(confirm_msg):
        log.info("Aborted by user")
        sys.exit(0)

    switch_to_project(dst_project)

    if watch:
        event_watcher = start_event_watcher(dst_project)

    DeployRunner(
        template_dir,
        dst_project,
        env_config_handler,
        ignore_requires=ignore_requires,
        service_sets_selected=sets_selected,
        resources_scale_factor=scale_resources,
        root_custom_dir=root_custom_dir,
        specific_components=specific_components,
        label=label,
        skip=skip.split(",") if skip else None,
        dry_run=False,
        concurrent=concurrent,
        threadpool_size=threadpool_size,
    ).run()

    if watch and event_watcher:
        event_watcher.stop()

    list_routes(dst_project)


@main.command("wipe", help="Delete everything from project")
@click.option("--no-confirm", "-f", is_flag=True, help="Do not prompt for confirmation")
@click.option(
    "--label", "-l", default=None, help="Delete only a specific label.  E.g. '-l app=test'"
)
@click.argument("dst_project")
def wipe_project(no_confirm, dst_project, label):
    verify_label(label)
    return wipe(no_confirm, dst_project, label)


@main.command("list-routes", help="List routes currently in the project")
@click.argument("dst_project")
@output_option
def list_act_routes(dst_project, output):
    return list_routes(dst_project, output)


@main.command("list-sets", help="List service sets available in template dir")
@click.option("--template-dir", "-t", default=None, help="Template directory (default 'templates')")
@output_option
def list_act_sets(template_dir, output):
    template_dir = get_dir(template_dir, "templates", "template")
    return list_sets(template_dir, output)


if __name__ == "__main__":
    main()



from __future__ import print_function

import glob
import json
import logging
import sys
import threading
import time
import os
import yaml
import re
from functools import reduce

from anytree import Node, RenderTree, PreOrderIter
import sh
from sh import ErrorReturnCode, TimeoutException
from wait_for import wait_for, TimedOutError

log = logging.getLogger(__name__)

# Resource types and their cli shortcuts
# Mostly listed here: https://docs.openshift.com/online/cli_reference/basic_cli_operations.html
SHORTCUTS = {
    "configmap": "cm",
    "cronjob": "cj",
    "build": None,
    "buildconfig": "bc",
    "daemonset": "ds",
    "deployment": "deploy",
    "deploymentconfig": "dc",
    "event": "ev",
    "imagestream": "is",
    "imagestreamtag": "istag",
    "imagestreamimage": "isimage",
    "job": None,
    "limitrange": "limits",
    "node": "no",
    "pod": "po",
    "resourcequota": "quota",
    "replicationcontroller": "rc",
    "secrets": "secret",
    "service": "svc",
    "serviceaccount": "sa",
    "statefulset": "sts",
    "persistentvolume": "pv",
    "persistentvolumeclaim": "pvc",
    "replicaset": "rs",
    "route": None,
}


INVALID_RESOURCE_REGEX = re.compile(
    r'The (\S+) "(\S+)" is invalid: metadata.resourceVersion: Invalid value: 0x0'
)


def abort():
    log.error("Hit fatal error!  Aborting.")
    sys.exit(1)


def validate_list_of_strs(item_name, section, list_):
    bad = False

    try:
        iter(list_)
    except TypeError:
        bad = True
    else:
        if not all([isinstance(i, str) for i in list_]):
            bad = True

    if bad:
        raise ValueError(f"'{item_name}' in '{section}' is not a list of strings")


def object_merge(old, new, merge_lists=True):
    """
    Recursively merge two data structures

    Thanks rsnyman :)
    https://github.com/rochacbruno/dynaconf/commit/458ffa6012f1de62fc4f68077f382ab420b43cfc#diff-c1b434836019ae32dc57d00dd1ae2eb9R15
    """
    if isinstance(old, list) and isinstance(new, list) and merge_lists:
        for item in old[::-1]:
            new.insert(0, item)
    if isinstance(old, dict) and isinstance(new, dict):
        for key, value in old.items():
            if key not in new:
                new[key] = value
            else:
                object_merge(value, new[key])
    return new


def traverse_keys(d, keys, default=None):
    """
    Allows you to look up a 'path' of keys in nested dicts without knowing whether each key exists
    """
    key = keys.pop(0)
    item = d.get(key, default)
    if len(keys) == 0:
        return item
    if not item:
        return default
    return traverse_keys(item, keys, default)


def parse_restype(string):
    """
    Given a resource type or its shortcut, return the full resource type name.
    """
    string_lower = string.lower()
    if string_lower in SHORTCUTS:
        return string_lower

    for resource_name, shortcut in SHORTCUTS.items():
        if string_lower == shortcut:
            return resource_name

    raise ValueError("Unknown resource type: {}".format(string))


def get_cfg_files_in_dir(path):
    """
    Get a list of all .yml/.json files in a dir

    Ignore the special _cfg file
    """
    files = list(glob.glob(os.path.join(path, "*.yaml")))
    files.extend(list(glob.glob(os.path.join(path, "*.yml"))))
    files.extend(list(glob.glob(os.path.join(path, "*.json"))))
    return [f for f in files if not os.path.basename(f).startswith("_cfg")]


def load_cfg_file(path):
    if not os.path.isfile(path):
        raise ValueError("Path '{}' is not a file or does not exist".format(path))

    _, file_ext = os.path.splitext(path)

    with open(path, "rb") as f:
        if file_ext == ".yaml" or file_ext == ".yml":
            content = yaml.safe_load(f)
        elif file_ext == ".json":
            content = json.load(f)
        else:
            raise ValueError("File '{}' must be a YAML or JSON file".format(path))

    if not content:
        raise ValueError("File '{}' is empty!".format(path))

    return content


def get_dir(value, default_value, dir_type, optional=False):
    path = value or default_value
    required_dir_does_not_exist = not optional and not os.path.exists(path)
    required_dir_is_not_a_dir = not optional and not os.path.isdir(path)
    path_exists_but_not_a_dir = os.path.exists(path) and not os.path.isdir(path)
    if required_dir_does_not_exist:
        log.error("%s directory missing: %s", dir_type, path)
        abort()
    if required_dir_is_not_a_dir or path_exists_but_not_a_dir:
        log.error("%s directory invalid: %s", dir_type, path)
        abort()
    path = os.path.abspath(path)
    if os.path.exists(path):
        log.info("Found %s path: %s", dir_type, path)
    return path


def all_sets(template_dir):
    try:
        cfg_data = load_cfg_file(f"{template_dir}/_cfg.yaml")
    except ValueError:
        try:
            cfg_data = load_cfg_file(f"{template_dir}/_cfg.yml")
        except ValueError as err:
            log.error("Error: template dir '%s' invalid: %s", template_dir, str(err))
            abort()

    try:
        stages = cfg_data["deploy_order"]
    except KeyError:
        log.error("Error: template dir '%s' invalid: _cfg file has no 'deploy_order'")
        abort()

    sets = reduce(lambda acc, s: acc + s.get("components", []), stages.values(), [])

    return sets


def _only_immutable_errors(err_lines):
    return all("field is immutable after creation" in line.lower() for line in err_lines)


def _conflicts_found(err_lines):
    return any("error from server (conflict)" in line.lower() for line in err_lines)


def _get_logging_args(args, kwargs):
    # Format the cmd args/kwargs for log printing before the command is run
    # Maybe 'sh' provides an easy way to do this...?
    cmd_args = " ".join([str(arg) for arg in args if arg is not None])

    cmd_kwargs = []
    for key, val in kwargs.items():
        if key.startswith("_"):
            continue
        if len(key) > 1:
            cmd_kwargs.append("--{} {}".format(key, val))
        else:
            cmd_kwargs.append("-{} {}".format(key, val))
    cmd_kwargs = " ".join(cmd_kwargs)

    return cmd_args, cmd_kwargs


def _exec_oc(*args, **kwargs):
    _silent = kwargs.pop("_silent", False)
    _ignore_immutable = kwargs.pop("_ignore_immutable", True)
    _retry_conflicts = kwargs.pop("_retry_conflicts", True)
    _stdout_log_prefix = kwargs.pop("_stdout_log_prefix", " |stdout| ")
    _stderr_log_prefix = kwargs.pop("_stderr_log_prefix", " |stderr| ")

    kwargs["_bg"] = True
    kwargs["_bg_exc"] = False

    err_lines = []
    out_lines = []

    def _err_line_handler(line, _, process):
        threading.current_thread().name = f"pid-{process.pid}"
        log.info("%s%s", _stderr_log_prefix, line.rstrip())
        err_lines.append(line)

    def _out_line_handler(line, _, process):
        threading.current_thread().name = f"pid-{process.pid}"
        if not _silent:
            log.info("%s%s", _stdout_log_prefix, line.rstrip())
        out_lines.append(line)

    retries = 3
    last_err = None
    for count in range(1, retries + 1):
        cmd = sh.oc(*args, **kwargs, _tee=True, _out=_out_line_handler, _err=_err_line_handler)
        if not _silent:
            cmd_args, cmd_kwargs = _get_logging_args(args, kwargs)
            log.info("running (pid %d): oc %s %s", cmd.pid, cmd_args, cmd_kwargs)
        try:
            return cmd.wait()
        except ErrorReturnCode as err:
            # Sometimes stdout/stderr is empty in the exception even though we appended
            # data in the callback. Perhaps buffers are not being flushed ... so just
            # set the out lines/err lines we captured on the Exception before re-raising it by
            # re-init'ing the err and causing it to rebuild its message template.
            #
            # see https://github.com/amoffat/sh/blob/master/sh.py#L381
            err.__init__(
                full_cmd=err.full_cmd,
                stdout="\n".join(out_lines).encode(),
                stderr="\n".join(err_lines).encode(),
                truncate=err.truncate,
            )

            # Make these plain strings for easier exception handling
            err.stdout = "\n".join(out_lines)
            err.stderr = "\n".join(err_lines)

            last_err = err

            # Ignore warnings that are printed to stderr in our error analysis
            err_lines = [line for line in err_lines if not line.lstrip().startswith("Warning:")]

            # Check if these are errors we should handle
            if _ignore_immutable and _only_immutable_errors(err_lines):
                log.warning("Ignoring immutable field errors")
                break
            elif _retry_conflicts and _conflicts_found(err_lines):
                log.warning(
                    "Hit resource conflict, retrying in 1 sec (attempt %d/%d)", count, retries
                )
                time.sleep(1)
                continue

            # Bail if not
            raise
    else:
        log.error("Retried %d times, giving up", retries)
        raise last_err


def oc(*args, **kwargs):
    """
    Run 'sh.oc' and print the command, show output, catch errors, etc.

    Optional kwargs:
        _reraise: if ErrorReturnCode is hit, don't exit, re-raise it
        _exit_on_err: sys.exit(1) if this command fails (default True)
        _silent: don't print command or resulting stdout (default False)
        _ignore_immutable: ignore errors related to immutable objects (default True)
        _retry_conflicts: retry commands if a conflict error is hit
        _stdout_log_prefix: prefix this string to stdout log output (default " |stdout| ")
        _stderr_log_prefix: prefix this string to stderr log output (default " |stderr| ")

    Returns:
        None if cmd fails and _exit_on_err is False
        command output (str) if command succeeds
    """
    _exit_on_err = kwargs.pop("_exit_on_err", True)
    _reraise = kwargs.pop("_reraise", False)
    # The _silent/_ignore_immutable/_retry_conflicts kwargs are passed on so don't pop them yet

    try:
        return _exec_oc(*args, **kwargs)
    except ErrorReturnCode:
        if _reraise:
            raise
        elif _exit_on_err:
            abort()
        else:
            if not kwargs.get("_silent"):
                log.warning("Non-zero return code ignored")


def apply_template(project, template):
    try:
        oc("apply", "-f", "-", "-n", project, _in=template.dump_processed_json(), _reraise=True)
    except ErrorReturnCode as err:
        # Work-around for resourceVersion errors.
        # See https://www.timcosta.io/kubernetes-service-invalid-clusterip-or-resourceversion/
        matches = INVALID_RESOURCE_REGEX.findall(err.stderr)
        if matches:
            for restype, name in matches:
                restype = restype.rstrip("s")  # remove plural language
                if template.get_processed_item(restype, name):  # ensure we sent this item's config
                    log.warning(
                        "Removing last-applied-configuration annotation from %s/%s", restype, name
                    )
                    oc(
                        "annotate",
                        restype,
                        name,
                        "kubectl.kubernetes.io/last-applied-configuration-",
                    )
            oc("apply", "-f", "-", "-n", project, _in=template.dump_processed_json())
        else:
            abort()


def switch_to_project(project):
    try:
        oc("get", "project", project, _reraise=True)
    except ErrorReturnCode:
        log.error("Unable to get project '%s', trying to create it...", project)
        oc("new-project", project, _exit_on_err=True)
    oc("project", project, _exit_on_err=True)


def get_json(restype, name=None, label=None):
    """
    Run 'oc get' for a given resource type/name/label and return the json output.

    If name is None all resources of this type are returned

    If label is not provided, then "oc get" will not be filtered on label
    """
    restype = parse_restype(restype)

    args = ["get", restype]
    if name:
        args.append(name)
    if label:
        args.extend(["-l", label])
    try:
        output = oc(*args, o="json", _exit_on_err=False, _silent=True)
    except ErrorReturnCode as err:
        if "NotFound" in err.stderr:
            return {}

    try:
        parsed_json = json.loads(str(output))
    except ValueError:
        return {}

    return parsed_json


def rollout(dc_name):
    """Rollout a deployment, wait for new revision to start deploying, wait for it to go active."""

    def _get_revision():
        try:
            return get_json("dc", dc_name)["status"]["latestVersion"]
        except KeyError:
            return "error"

    old_revision = _get_revision()
    try:
        oc("rollout", "latest", "dc/{}".format(dc_name), _reraise=True)
    except ErrorReturnCode as err:
        if "is already in progress" in str(err):
            pass
    else:
        # Wait for the new revision to start deploying
        for _ in range(0, 60):
            if _get_revision() != old_revision:
                break
            log.info("Waiting for rollout on dc/%s to begin", dc_name)
            time.sleep(1)

    wait_for_ready("dc", dc_name)


def get_routes():
    """
    Get all routes in the project.

    Return dict with key of service name, value of http route
    """
    data = get_json("route")
    ret = {}
    for route in data.get("items", []):
        ret[route["metadata"]["name"]] = route["spec"]["host"]
    return ret


class StatusError(Exception):
    pass


def _check_status_for_restype(restype, json_data):
    """
    Depending on the resource type, check that it is "ready" or "complete"

    Uses the status json from an 'oc get'

    Returns True if ready, False if not.
    """
    try:
        status = json_data["status"]
    except KeyError:
        status = None

    try:
        name = json_data["metadata"]["name"]
    except KeyError:
        name = "unknown"

    if not status:
        return False

    restype = parse_restype(restype)

    if restype == "deploymentconfig" or restype == "deployment":
        spec_replicas = json_data["spec"]["replicas"]
        available_replicas = status.get("availableReplicas", 0)
        updated_replicas = status.get("updatedReplicas", 0)
        unavailable_replicas = status.get("unavailableReplicas", 1)
        if unavailable_replicas == 0:
            if available_replicas == spec_replicas and updated_replicas == spec_replicas:
                return True

    elif restype == "statefulset":
        spec_replicas = json_data["spec"]["replicas"]
        ready_replicas = status.get("readyReplicas", 0)
        return ready_replicas == spec_replicas

    elif restype == "daemonset":
        desired = status.get("desiredNumberScheduled", 1)
        available = status.get("numberAvailable")
        return desired == available

    elif restype == "pod":
        if status.get("phase").lower() == "running":
            return True

    elif restype == "build":
        phase = status.get("phase").lower()
        if phase == "cancelled":
            log.warning("Build '%s' was cancelled!", name)
            return True
        elif phase in ["completed", "complete"]:
            return True
        elif phase in ["failed", "error"]:
            raise StatusError("Build '{}' failed!".format(name))

    elif restype == "buildconfig":
        try:
            last_version = json_data["status"]["lastVersion"]
        except KeyError:
            log.debug("No builds triggered yet for 'bc/%s'", name)
            return False

        build = "{}-{}".format(name, last_version)
        log.debug("checking 'bc/%s' last triggered build: %s", name, build)
        return _check_status_for_restype("build", get_json("build", build))

    else:
        raise ValueError(
            "Checking status for resource type {} is not supported right now".format(restype)
        )


def _wait_with_periodic_status_check(timeout, key, restype, name):
    """Check if resource is ready using _check_status_for_restype, periodically log an update."""
    time_last_logged = time.time()
    time_remaining = timeout

    def _ready():
        nonlocal time_last_logged, time_remaining

        j = get_json(restype, name)
        if _check_status_for_restype(restype, j):
            return True

        if time.time() > time_last_logged + 60:
            time_remaining -= 60
            if time_remaining:
                log.info("[%s] waiting %dsec longer", key, time_remaining)
                time_last_logged = time.time()
        return False

    wait_for(
        _ready, timeout=timeout, delay=5, message="wait for '{}' to be ready".format(key),
    )


def wait_for_exists(restype, name, timeout=300):
    restype = parse_restype(restype)
    key = "{}/{}".format(SHORTCUTS.get(restype) or restype, name)
    log.info("[%s] waiting up to %dsec for resource to exist", key, timeout)

    def _exists():
        return bool(get_json(restype, name))

    wait_for(_exists, timeout=timeout, delay=5, message="wait for '{}' to exist".format(key))


def wait_for_ready(restype, name, timeout=300, exit_on_err=False, _result_dict=None):
    """
    Wait {timeout} for resource to be complete/ready/active.

    Args:
        restype: type of resource, which can be "build", "dc", "deploymentconfig"
        name: name of resource
        timeout: time in secs to wait for resource to become ready
        exit_on_err: if resource fails to become ready, exit with error code

    Returns:
        True if ready,
        False if timed out

    '_result_dict' can be passed when running this in a threaded fashion
    to store the result of this wait as:
        _result_dict[resource_name] = True or False
    """
    restype = parse_restype(restype)
    key = "{}/{}".format(SHORTCUTS.get(restype) or restype, name)

    if _result_dict is None:
        _result_dict = dict()
    _result_dict[key] = False

    log.info("[%s] waiting up to %dsec for resource to be ready", key, timeout)

    try:
        # Do not use rollout status for statefulset/daemonset yet until we can handle
        # https://github.com/kubernetes/kubernetes/issues/64500
        if restype in ["deployment", "deploymentconfig"]:
            # use oc rollout status for the applicable resource types
            oc(
                "rollout",
                "status",
                key,
                _reraise=True,
                _timeout=timeout,
                _stdout_log_prefix=f"[{key}] ",
                _stderr_log_prefix=f"[{key}]  ",
            )
        else:
            _wait_with_periodic_status_check(timeout, key, restype, name)

        log.info("[%s] is ready!", key)
        _result_dict[key] = True
        return True
    except (StatusError, ErrorReturnCode) as err:
        log.error("[%s] hit error waiting for resource to be ready: %s", key, str(err))
    except (TimeoutException, TimedOutError):
        log.error("[%s] timed out waiting for resource to be ready", key)
    if exit_on_err:
        abort()
    return False


def wait_for_ready_threaded(restype_name_list, timeout=300, exit_on_err=False):
    """
    Wait for multiple delpoyments in a threaded fashion.

    Args:
        restype_name_list: list of tuples with (resource_type, resource_name,)
        timeout: timeout for each thread
        exit_on_err: when all threads finish, if any failed, exit

    Returns:
        True if all deployments are ready
        False if any failed
    """
    result_dict = dict()
    threads = [
        threading.Thread(target=wait_for_ready, args=(restype, name, timeout, False, result_dict))
        for restype, name in restype_name_list
    ]
    for thread in threads:
        thread.daemon = True
        thread.name = thread.name.lower()  # because I'm picky
        thread.start()
    for thread in threads:
        thread.join()

    failed = [key for key, result in result_dict.items() if not result]

    if failed:
        log.info("Some resources failed to become ready: %s", ", ".join(failed))
        if exit_on_err:
            abort()
        return False
    return True


def any_pods_running(dc_name):
    """
    Return true if any pods are running in the deployment config
    """
    pod_data = get_json("pod", label="deploymentconfig={}".format(dc_name))
    if not pod_data or not len(pod_data.get("items", [])):
        log.info("No pods found for dc %s", dc_name)
        return False
    for pod in pod_data["items"]:
        if _check_status_for_restype("pod", pod):
            return True
    return False


def all_pods_running(dc_name):
    """
    Return true if all pods are running in the deployment config
    """
    pod_data = get_json("pod", label="deploymentconfig={}".format(dc_name))
    if not pod_data or not len(pod_data.get("items", [])):
        log.info("No pods found for dc %s", dc_name)
        return False
    statuses = []
    for pod in pod_data["items"]:
        statuses.append(_check_status_for_restype("pod", pod))
    return len(statuses) and all(statuses)


def no_pods_running(dc_name):
    """
    Return true if there are no pods running in the deployment
    """
    return not any_pods_running(dc_name)


def stop_deployment(dc_name, timeout=180):
    """
    Pause a deployment, delete all of its replication controllers, wait for all pods to shut down
    """
    if not any_pods_running(dc_name):
        log.info("No pods running for dc '%s', nothing to stop", dc_name)
        return

    log.info("Patching deployment config for '%s' to pause rollouts", dc_name)
    try:
        oc("rollout", "pause", "dc/{}".format(dc_name), _reraise=True)
    except sh.ErrorReturnCode as err:
        if "is already paused" in str(err.stderr):
            pass

    log.info("Removing replication controllers for '%s'", dc_name)
    rc_data = get_json("rc", label="openshift.io/deployment-config.name={}".format(dc_name))
    if not rc_data or not len(rc_data.get("items", [])):
        raise Exception("Unable to find replication controllers for '{}'".format(dc_name))
    for rc in rc_data["items"]:
        rc_name = rc["metadata"]["name"]
        oc("delete", "rc", rc_name)

    log.info("Waiting for pods related to '%s' to terminate", dc_name)
    wait_for(
        no_pods_running,
        func_args=(dc_name,),
        message="wait for deployment '{}' to be terminated".format(dc_name),
        timeout=timeout,
        delay=5,
        log_on_loop=True,
    )


def dc_ready(dc_name):
    dc_json = get_json("dc", dc_name)
    return _check_status_for_restype("dc", dc_json) and all_pods_running(dc_name)


def start_deployment(dc_name, timeout=180):
    if dc_ready(dc_name):
        log.info("Deployment '%s' already deployed and running, skipping deploy for it", dc_name)
        return

    log.info("Patching deployment config for '%s' to resume rollouts", dc_name)
    try:
        oc("rollout", "resume", "dc/{}".format(dc_name), _reraise=True)
    except sh.ErrorReturnCode as err:
        if "is not paused" in str(err.stderr):
            pass

    log.info("Triggering new deploy for '%s'", dc_name)
    try:
        oc("rollout", "latest", "dc/{}".format(dc_name), _reraise=True)
    except sh.ErrorReturnCode as err:
        if "already in progress" in str(err.stderr):
            pass

    log.info("Waiting for pod related to '%s' to finish deploying", dc_name)
    wait_for(
        dc_ready,
        func_args=(dc_name,),
        message="wait for deployment '{}' to be ready".format(dc_name),
        delay=5,
        timeout=timeout,
        log_on_loop=True,
    )


def get_server_info():
    """Return server connected on"""
    return oc("whoami", "--show-server", _silent=True)


def cancel_builds(bc_name):
    oc(
        "cancel-build",
        f"bc/{bc_name}",
        state="new,pending,running",
        _timeout=120,
        _exit_on_err=False,
    )

    # Check if there's any lingering builds
    builds = get_json("build", label=f"openshift.io/build-config.name={bc_name}")
    lingering_builds = []
    for build in builds.get("items", []):
        # delete these builds rather than cancelling them, since jenkins pipeline builds
        # can remain stuck in certain states if OpenShift Sync plugin is broken
        status = build.get("status") or {}
        phase = status.get("phase", "").lower()
        if phase in ["new", "pending"]:
            build_name = build["metadata"]["name"]
            lingering_builds.append(build_name)

    if lingering_builds:
        log.warning("Found lingering builds for bc/%s which will be deleted", bc_name)
        for build_name in lingering_builds:
            oc("delete", "build", build_name)


def get_input_image(buildconfig, trigger):
    """
    Look up the image stream input image for a build triggered by imagechange.
    """
    bc = buildconfig
    input_image = None

    if trigger.get("imageChange", {}) != {}:
        # the image used for trigger is explicitly defined, we're done
        input_image = trigger["imageChange"]["from"]["name"]
        return input_image

    # we need to look up the image used for trigger in the bc's configuration
    # check if there's a dockerfile with FROM line
    dockerfile = traverse_keys(bc, ["spec", "source", "dockerfile"])
    if dockerfile:
        for line in dockerfile.splitlines():
            if line.startswith("FROM"):
                input_image = line.split()[1]
                if ":" not in input_image:
                    input_image = f"{input_image}:latest"
                break

    # check if the source imagestreamtag is defined in the strategy config
    for key in ["dockerStrategy", "sourceStrategy", "customStrategy"]:
        from_kind = traverse_keys(bc, ["spec", "strategy", key, "from", "kind"], "").lower()
        if from_kind == "imagestreamtag":
            input_image = bc["spec"]["strategy"][key]["from"]["name"]

    return input_image


def get_build_tree(buildconfigs):
    """
    Analyze build configurations to find which builds are 'linked'.

    Linked builds are those which output to an ImageStream that another BuildConfig then
    uses as its 'from' image.

    Returns a list of lists where item 0 in each list is the parent build and the items following
    it are all child build configs that will be fired at some point after the parent completes
    """
    bcs_using_input_image = {}
    bc_creating_output_image = {None: None}
    node_for_bc = {}
    for bc in buildconfigs:
        bc_name = bc["metadata"]["name"]
        node_for_bc[bc_name] = Node(bc_name)

        # look up output image
        if traverse_keys(bc, ["spec", "output", "to", "kind"], "").lower() == "imagestreamtag":
            output_image = bc["spec"]["output"]["to"]["name"]
            bc_creating_output_image[output_image] = bc_name

        # look up input image
        for trigger in traverse_keys(bc, ["spec", "triggers"], []):
            if trigger.get("type", "").lower() == "imagechange":
                input_image = get_input_image(bc, trigger)
                if input_image not in bcs_using_input_image:
                    bcs_using_input_image[input_image] = []
                bcs_using_input_image[input_image].append(bc_name)

    # attach each build to its parent build
    for input_image, bc_names in bcs_using_input_image.items():
        for bc_name in bc_names:
            parent_bc = bc_creating_output_image.get(input_image)
            if parent_bc:
                node_for_bc[bc_name].parent = node_for_bc[parent_bc]

    rendered_trees = []
    root_nodes = [n for _, n in node_for_bc.items() if n.is_root]
    for root_node in root_nodes:
        for pre, _, node in RenderTree(root_node):
            rendered_trees.append(f"  {pre}{node.name}")

    if rendered_trees:
        log.info("build config tree:\n\n%s", "\n".join(rendered_trees))

    return [[node.name for node in PreOrderIter(root_node)] for root_node in root_nodes]


def get_next_build(bc_name):
    """
    Return the upcoming build name we can expect to see for a given build config
    """
    json_data = get_json("bc", bc_name)
    last_version = traverse_keys(json_data, ["status", "lastVersion"], 0)
    next_build = "{}-{}".format(bc_name, last_version + 1)
    return next_build


def trigger_builds(buildconfigs):
    """
    Trigger parent build configs based on a build tree and return the resources to wait for
    """
    bcs = buildconfigs

    builds_to_wait_for = []

    build_tree = get_build_tree(bcs)
    for sub_tree in build_tree:
        parent = sub_tree[0]
        for bc_name in sub_tree:
            # Cancel any new/pending builds
            cancel_builds(bc_name)
            builds_to_wait_for.append(("build", get_next_build(bc_name)))
        log.info("triggering build for '%s'", parent)
        oc("start-build", "bc/{}".format(parent))

    return builds_to_wait_for



import pytest


def _is_test_path(path):
    return "TEST" in path


def _patch_exists(path):
    if _is_test_path(path):
        print(f"Overriding os.path.exists for path={path}")
        return True


def _patch_isdir(path):
    if _is_test_path(path):
        print(f"Overriding os.path.isdir for path={path}")
        return True


def _patch_isfile(path):
    if _is_test_path(path):
        print(f"Overriding os.path.isfile for path={path}")
        return True


@pytest.fixture
def patch_os_path(monkeypatch):
    monkeypatch.setattr("os.path.exists", _patch_exists)
    monkeypatch.setattr("os.path.isdir", _patch_isdir)
    monkeypatch.setattr("os.path.isfile", _patch_isfile)



import ocdeployer.utils as utils


def test_object_merge_merges_lists():
    edit = [1, 3]
    add = [2]

    utils.object_merge(add, edit)

    assert edit == [2, 1, 3]


def test_obect_merge_merges_dicts():
    edit = {"thing": "stuff"}
    add = {"test": "add"}

    utils.object_merge(add, edit)

    assert edit == {"thing": "stuff", "test": "add"}


def test_object_merge_keeps_original_dict_keys():
    edit = {"thing": 3}
    add = {"thing": 1}

    utils.object_merge(add, edit)

    assert edit == {"thing": 3}


def test_object_merge_recursively_merges_dicts():
    edit = {"things": {"stuff": 1}}
    add = {"things": {"new": 2}}

    utils.object_merge(add, edit)

    assert edit == {"things": {"stuff": 1, "new": 2}}



import pytest

from ocdeployer.templates import Template


@pytest.mark.parametrize(
    "value,expected",
    (
        (True, "true"),
        ("True", "True"),
        ("true", "true"),
        ("123", "123"),
        (123, "123"),
        ("123:123:123", "123:123:123"),
        ("some text", "some text"),
    ),
)
def test_template_oc_param_format(value, expected):
    assert Template._format_oc_parameter(value) == expected



import logging

from .utils import oc, get_json, validate_list_of_strs


log = logging.getLogger("ocdeployer.images")


# Template of an ImageStream config that 'oc import-image' will create
def _get_is_config(is_name, is_tag, image_from, scheduled):
    return {
        "apiVersion": "image.openshift.io/v1",
        "kind": "ImageStream",
        "metadata": {"name": is_name},
        "spec": {
            "lookupPolicy": {"local": False},
            "tags": [
                {
                    "from": {"kind": "DockerImage", "name": image_from},
                    "importPolicy": {"scheduled": scheduled},
                    "name": is_tag,
                    "referencePolicy": {"type": "Source"},
                }
            ],
        },
    }


def _parse_istag(istag):
    """Append "latest" tag onto istag if it has no tag."""
    if ":" not in istag:
        return f"{istag}:latest"
    return istag


def _parse_new_style(images):
    """Handles new-style images config syntax."""
    parsed_images = []

    for img in images:
        if not isinstance(img, dict):
            raise ValueError("entries in 'images' must be of type 'dict'")
        if len(img.keys()) >= 2 and all([k in img for k in ["istag", "from"]]):
            # This entry is using long-style image definition, e.g.
            #   images:
            #   - istag: "image_name"
            #   - from: "quay.io/some/image_name:image_tag"
            #   - envs: ["stage", "prod"]
            istag = _parse_istag(img["istag"])
            _from = img["from"]
            envs = img.get("envs", [])
            scheduled = img.get("scheduled", True)
        elif len(img.keys()) == 1 and all(
            [k not in img for k in ["istag", "from", "envs", "scheduled"]]
        ):
            # This entry is using short-style image definition, e.g.
            #   images:
            #   - "image_name:image_tag": "quay.io/some/image_name:image_tag"
            istag, _from = list(img.items())[0]
            istag = _parse_istag(istag)
            scheduled = True
            envs = []
        else:
            raise ValueError("Unknown syntax for 'images' section of config")

        if not isinstance(istag, str) or not isinstance(_from, str):
            raise ValueError("'istag' and 'from' must be a of type 'string'")
        validate_list_of_strs("envs", "images", envs)

        parsed_images.append({"istag": istag, "from": _from, "envs": envs, "scheduled": scheduled})

    return parsed_images


def _parse_old_style(images):
    """Handles old-style images config.

    e.g.:

    images:
        istag1: "docker.io/from-uri"
        "istag2:latest": "fedora:latest"
    """
    parsed_images = []

    for istag, _from in images.items():
        if not isinstance(istag, str) or not isinstance(_from, str):
            raise ValueError("keys and values in 'images' must be a of type 'string'")
        parsed_images.append(
            {"istag": _parse_istag(istag), "from": _from, "envs": [], "scheduled": True}
        )

    return parsed_images


def parse_config(config):
    if "images" in config:
        if isinstance(config["images"], dict):
            return _parse_old_style(config["images"])
        elif isinstance(config["images"], list):
            return _parse_new_style(config["images"])
    return []


class ImageImporter:
    """
    A singleton which handles importing images

    Keeps track of which secrets have been imported so we don't keep re-importing.
    """

    imported_istags = []

    @classmethod
    def _retag_image(cls, istag, image_from, scheduled):
        oc(
            "tag", f"--scheduled={scheduled}", "--source=docker", image_from, istag,
        )
        oc("import-image", istag)

    @classmethod
    def _import_image(cls, istag, image_from, scheduled):
        oc(
            "import-image",
            istag,
            "--from={}".format(image_from),
            "--confirm",
            f"--scheduled={scheduled}",
            _reraise=True,
        )

    @classmethod
    def do_import(cls, istag, image_from, scheduled, **kwargs):
        if istag in cls.imported_istags:
            log.warning("istag '%s' already imported, skipping repeat import...", istag)
            return

        scheduled = "True" if scheduled else "False"
        if get_json("istag", istag):
            cls._retag_image(istag, image_from, scheduled)
        else:
            cls._import_image(istag, image_from, scheduled)

        cls.imported_istags.append(istag)


def _get_args(config, env_names):
    args = []
    images = parse_config(config)
    for img_data in images:
        istag = img_data["istag"]
        image_from = img_data["from"]
        scheduled = img_data.get("scheduled", True)
        if not img_data["envs"] or any([e in env_names for e in img_data["envs"]]):
            args.append((istag, image_from, scheduled))
        else:
            log.info("Skipping import of image '%s', not enabled for this env", img_data["istag"])
    return args


def import_images(config, env_names):
    """Import the specified images listed in a _cfg.yml"""
    for args in _get_args(config, env_names):
        ImageImporter.do_import(*args)


def get_is_configs(config, env_names):
    """Return a list of image stream configs that would be created based on images in _cfg.yml"""
    items = {}
    for args in _get_args(config, env_names):
        istag, image_from, scheduled = args
        is_name, is_tag = istag.split(":")
        items[istag] = _get_is_config(is_name, is_tag, image_from, scheduled)
    return items



import logging
import threading

import datetime
import pytz
from kubernetes import client, config, watch


log = logging.getLogger(__name__)


class EventWatcher(threading.Thread):
    def __init__(self, namespace, *args, **kwargs):
        daemon = kwargs.pop("daemon", True)
        super().__init__(*args, **kwargs, daemon=daemon)
        # Since we have already run 'oc project', our kube config has auth info
        config.load_kube_config()
        self._watcher = None
        self.namespace = namespace
        self.name = "event_watcher"
        self.v1_client = client.CoreV1Api()

    def get_all_events(self, _events=None, _continue=None):
        """
        Retrieve all events from the given namespace

        Uses _continue functionality recursively
        """
        if not _events:
            _events = []

        if _continue:
            v1_event_list = self.v1_client.list_namespaced_event(
                self.namespace, _continue=_continue
            )
        else:
            v1_event_list = self.v1_client.list_namespaced_event(self.namespace)

        _events.extend(v1_event_list.items)
        if v1_event_list.metadata._continue:
            self.get_all_events(_events, v1_event_list.metadata._continue)

        return _events

    def run(self):
        """
        Log all events generated for a namespace

        Will only log events that were generated after the EventWatcher thread has started
        """
        log.info("Starting event watcher on namespace '%s'", self.namespace)

        # Get the events from the API and grab the last one listed (if there is one)
        # We only want to show events that appear AFTER this event.
        # This is similar in function to 'oc get events --watch-only=true'
        old_event_time = pytz.utc.localize(datetime.datetime.min)
        old_events = self.get_all_events()
        if old_events:
            last_old_event = sorted(old_events, key=lambda x: x.last_timestamp)[-1]
            old_event_time = last_old_event.last_timestamp

        self._watcher = watch.Watch()
        last_event_info = None
        for event in self._watcher.stream(self.v1_client.list_namespaced_event, self.namespace):
            obj = event["object"]
            event_info = (
                obj.last_timestamp,
                obj.involved_object.kind,
                obj.involved_object.name,
                obj.type,
                obj.reason,
                obj.message,
            )
            # Only print new events, and don't print repeat events
            if obj.last_timestamp > old_event_time and event_info != last_event_info:
                log.info(" --> [%s] [%s %s] [%s - %s] %s", *event_info)
                last_event_info = event_info

    def stop(self):
        log.info("Stopping event watcher on namespace '%s'", self.namespace)
        self._watcher.stop()


def start_event_watcher(namespace):
    try:
        event_watcher = EventWatcher(namespace)
    except Exception:
        log.exception("Failed to init event watcher, unable to monitor events")
        event_watcher = None

    if event_watcher:
        event_watcher.start()

    return event_watcher



"""
Helper methods for dealing with openshift template
"""
import os
import json
import logging
import re
import yaml

from cached_property import cached_property
from jinja2 import Template as Jinja2Template

from .utils import oc, parse_restype, get_cfg_files_in_dir


log = logging.getLogger(__name__)


PARAM_REGEX = re.compile(r"(^.*)(\$\{\{\S+\}\})(.*$)", re.MULTILINE)
RESOURCE_REGEX = re.compile(r"(\d+)(\.\d+)?([A-Za-z]+)?")


def get_templates_in_dir(path):
    """
    Given a directory path, returns a dict with keys: template name, vals: Template instance

    Note that template name is the file name minus extension
    """
    templates = [Template(path=file_path) for file_path in get_cfg_files_in_dir(path)]
    template_for_name = {os.path.basename(temp.file_name): temp for temp in templates}
    return template_for_name


def _scale_val(val, scale_factor):
    """
    Parse out the number from a kubernetes resource string and scale it by scale_factor

    Numbers are rounded to 1 decimal place.

    Examples:
    * "500Mi" scaled by .1 returns "50.0Mi"
    * "2" scaled by .5 returns "1"
    * "200m" scaled by 2 returns "400m"
    """
    match = RESOURCE_REGEX.match(str(val))
    if match:
        base_num, decimal, unit = match.groups()
    else:
        return val
    float_num = float("{}{}".format(base_num, decimal if decimal else ".0"))
    return "{}{}".format(str(round(float_num * scale_factor, 1)), unit if unit else "")


def _scale_limits_and_requests(data, scale_factor):
    for limit_key, limit_val in data.get("limits", {}).items():
        old_val = limit_val[0] if isinstance(limit_val, tuple) else limit_val
        new_val = _scale_val(old_val, scale_factor)
        data["limits"][limit_key] = new_val
        log.info("Adjusted limits for '%s', old: %s, new: %s", limit_key, old_val, new_val)
    for request_key, request_val in data.get("requests", {}).items():
        old_val = request_val[0] if isinstance(request_val, tuple) else request_val
        new_val = _scale_val(old_val, scale_factor)
        data["requests"][request_key] = new_val
        log.info("Adjusted requests for '%s', old: %s, new: %s", request_key, old_val, new_val)


def scale_resources(
    config_data, scale_factor, _current_dict_path=None, _obj_name=None, _obj_kind=None
):
    """
    Iterate through a processed config looking for any resources dictionaries.

    Scale the resource limits and requests by 'scale_factor' if any are present.
    """
    if not _current_dict_path:
        _current_dict_path = "items"
        config_data = config_data["items"]

    if isinstance(config_data, list):
        # Keep iterating thru lists
        for index, list_item in enumerate(config_data):
            scale_resources(
                list_item,
                scale_factor,
                "{}[{}]".format(_current_dict_path, index),
                _obj_name,
                _obj_kind,
            )
    elif isinstance(config_data, dict):
        # Keep iterating thru dicts
        _obj_name = config_data.get("metadata", {}).get("name") or _obj_name
        _obj_kind = config_data.get("kind") or _obj_kind
        for key, data in config_data.items():
            if key == "resources":
                # If we hit a 'resources' dict, scale it.
                if scale_factor <= 0:
                    log.info(
                        "Removing resource requests/limits for %s '%s' found at %s",
                        _obj_kind,
                        _obj_name,
                        _current_dict_path,
                    )
                    try:
                        del data["limits"]
                        del data["requests"]
                    except KeyError:
                        pass
                else:
                    log.info(
                        "Scaling resources for %s '%s' found at %s",
                        _obj_kind,
                        _obj_name,
                        _current_dict_path,
                    )
                    _scale_limits_and_requests(data, scale_factor)
            if isinstance(data, list) or isinstance(data, dict):
                scale_resources(
                    data,
                    scale_factor,
                    "{}['{}']".format(_current_dict_path, key),
                    _obj_name,
                    _obj_kind,
                )


class Template(object):
    """
    Represents an openshift template.

    A template can be processed multiple times with self.process()
    Each time it is processed, self.processed_content will get stored
    with the most recent result of that process
    """

    def __init__(self, path):
        """
        Constructor

        Args:
            path (str) -- path to file
            variables -- dict with key variable name/variable value for this template
        """
        self.path = path
        self.file_name, self.file_extension = os.path.splitext(self.path)
        self.processed_content = {}
        self.processed_jinja_content = {}

    @staticmethod
    def _jinja_safe(data):
        """
        Mark certain sections of text as "raw" for jinja processing

        Example:
        ${{PARAM}} is valid Open Shift template syntax, but by default jinja2
        would evalulate this as a variable and replace this to "$"
        """
        return PARAM_REGEX.sub(r"\1{% raw %}\2{% endraw %}\3", data)

    @cached_property
    def content(self):
        with open(self.path, "r") as f:
            return self._jinja_safe(f.read())

    @staticmethod
    def _format_oc_parameter(param_value):
        """
        Ensures that non-string parameter values are inserted into template properly
        """
        if isinstance(param_value, bool) or isinstance(param_value, int):
            return json.dumps(param_value)
        return param_value

    def _process_via_oc(self, content, parameters=None, label=None):
        """
        Run 'oc process' on the template and update content with the processed output

        Only passes in a parameter to 'oc process' if it is listed in the template's
        "parameters" config section.

        Args:
            parameters -- dict with key param name/param value for this template
            resources_scale_factor (float) -- scale any defined resource requests/limits by
                 this number (i.e., multiple their current values by this number)
            label (str) -- label to apply to all resources
        """
        if not parameters:
            parameters = {}

        params_and_vals = {}
        for param_name, param_value in parameters.items():
            params_and_vals[param_name] = "{}={}".format(
                param_name, self._format_oc_parameter(param_value)
            )

        extra_args = []
        # Only insert the parameter if it was defined in the template
        param_names_defined_in_template = [
            param.get("name") for param in content.get("parameters", [])
        ]
        skipped_params = []
        params_and_vals_used = []
        for param_name, string in params_and_vals.items():
            if param_name in param_names_defined_in_template:
                extra_args.extend(["-p", string])
                params_and_vals_used.append(string)
            else:
                skipped_params.append(param_name)

        log.info(
            "Running 'oc process' on template '%s' with parameters '%s'",
            self.file_name,
            ", ".join([string for string in params_and_vals_used]),
        )

        if skipped_params:
            log.info(
                "Skipped parameters defined in config but not present in template: %s",
                ", ".join(skipped_params),
            )

        if label:
            extra_args.extend(["-l", label])

        output = oc(
            "process",
            "--local",
            "-f",
            "-",
            "-o",
            "json",
            *extra_args,
            _silent=True,
            _in=json.dumps(content)
        )

        return json.loads(str(output))

    def _load_content(self, string):
        if self.path.endswith(".yml") or self.path.endswith(".yaml"):
            content = yaml.safe_load(string)
        else:
            content = json.loads(string)

        # Some checks to be (semi-)sure this is a valid template...
        if content.get("kind", "").lower() != "template":
            raise ValueError("Path '{}' is not of kind 'template'".format(self.path))
        if "objects" not in content:
            raise ValueError("Path '{}' has no 'objects'".format(self.path))

        return content

    def _process_via_jinja2(self, variables):
        log.info("Rendering template '%s' with jinja2", self.file_name)
        template = Jinja2Template(self.content)
        rendered_txt = template.render(**variables)
        if not rendered_txt.strip():
            log.info("Template '%s' is empty after jinja2 processing", self.file_name)
            self.processed_jinja_content = {}
        else:
            self.processed_jinja_content = self._load_content(rendered_txt)
        return self.processed_jinja_content

    def process_jinja(self, variables):
        return self._process_via_jinja2(variables)

    def process(self, variables, resources_scale_factor=1.0, label=None):
        # Run the template through jinja processing first
        jinjafied_content = self._process_via_jinja2(variables)

        # Once that is done, run it through standard openshift template processing
        if jinjafied_content:
            self.processed_content = self._process_via_oc(
                jinjafied_content, variables.get("parameters"), label
            )
            # Scale resources in the template
            if resources_scale_factor > 0 and resources_scale_factor != 1:
                log.info(
                    "Scaling resources for template '%s' by factor of %f",
                    self.file_name,
                    resources_scale_factor,
                )
                scale_resources(self.processed_content, resources_scale_factor)
        else:
            self.processed_content = {}

        return self.processed_content

    def dump_processed_json(self):
        return json.dumps(self.processed_content)

    def get_processed_items_for_restype(self, restype):
        """
        Return list of objects of type 'restype' in the processed template

        Note at the moment this only searches the 1st level of objects (for example,
        it will not return the name of a pod embedded within a deployment config)
        """
        restype = parse_restype(restype)

        return [
            obj for obj in self.processed_content.get("items", []) if obj["kind"].lower() == restype
        ]

    def get_processed_names_for_restype(self, restype):
        """
        Return list of names for all objects of type 'restype' in the processed template

        Note at the moment this only searches the 1st level of objects (for example,
        it will not return the name of a pod embedded within a deployment config)
        """
        return [obj["metadata"]["name"] for obj in self.get_processed_items_for_restype(restype)]

    def get_processed_item(self, restype, name):
        for obj in self.get_processed_items_for_restype(restype):
            if obj["metadata"]["name"] == name:
                return obj



from .secrets import parse_config as parse_secrets_config
from .images import parse_config as parse_images_config
from .utils import object_merge


def merge_list_of_dicts(old, new, key):
    """
    Merge a list of dictionary items based on a specific key.

    Dictionaries inside the list with a matching key get merged together.

    Assumes that a value for the given key is unique and appears only once.

    Example:
    list1 = [{"name": "one", "data": "stuff"}, {"name": "two", "data": "stuff2"}]
    list2 = [{"name": "one", "data": "newstuff"}]

    merge_list_of_dicts(list1, list2) returns:
    [{"name": "one", "data": "newstuff"}, {"name": "two", "data": "stuff2"}]
    """
    for old_item in reversed(old):
        matching_val = old_item[key]
        for new_item in new:
            if new_item[key] == matching_val:
                object_merge(old_item, new_item)
                break
        else:
            new.append(old_item)
    return new


def merge_cfgs(old, new):
    """
    Merges two _cfg definitions together.

    This is similar to an object_merge but with special processing of the 'images' and 'secrets'
    since these are lists of dictionaries

    object_merge simply merges items from an old & new list together. Instead, if 2 dictionaries
    items in the lists are defining data for the same image or secret, we want to merge the data of
    those dictionaries together.
    """
    old_images = parse_images_config(old)
    new_images = parse_images_config(new)
    old_secrets = parse_secrets_config(old)
    new_secrets = parse_secrets_config(new)

    # First do a standard object_merge, we'll then replace 'images' and 'secrets'
    object_merge(old, new)
    # Now merge the list of dictionaries together using their identifying key
    new["images"] = merge_list_of_dicts(old_images, new_images, key="istag")
    new["secrets"] = merge_list_of_dicts(old_secrets, new_secrets, key="name")

    return new



import pytest

from ocdeployer.images import ImageImporter, import_images


@pytest.fixture
def mock_oc(mocker):
    _mock_oc = mocker.patch("ocdeployer.images.oc")
    mocker.patch("ocdeployer.images.get_json", return_value={})
    yield _mock_oc


def _check_oc_calls(mocker, mock_oc):
    assert mock_oc.call_count == 2
    calls = [
        mocker.call(
            "import-image",
            "image1:tag",
            "--from=docker.url/image1:sometag",
            "--confirm",
            "--scheduled=True",
            _reraise=True,
        ),
        mocker.call(
            "import-image",
            "image2:tag",
            "--from=docker.url/image2:sometag",
            "--confirm",
            "--scheduled=True",
            _reraise=True,
        ),
    ]
    mock_oc.assert_has_calls(calls)


def test_images_short_style_syntax(mocker, mock_oc):
    config_content = {
        "images": [
            {"image1:tag": "docker.url/image1:sometag"},
            {"image2:tag": "docker.url/image2:sometag"},
        ]
    }
    ImageImporter.imported_istags = []
    import_images(config_content, [])

    _check_oc_calls(mocker, mock_oc)


def test_images_long_style_syntax(mocker, mock_oc):
    config_content = {
        "images": [
            {"istag": "image1:tag", "from": "docker.url/image1:sometag"},
            {"istag": "image2:tag", "from": "docker.url/image2:sometag"},
        ]
    }

    ImageImporter.imported_istags = []
    import_images(config_content, [])

    _check_oc_calls(mocker, mock_oc)


def test_images_old_style_syntax(mocker, mock_oc):
    config_content = {
        "images": {
            "image1:tag": "docker.url/image1:sometag",
            "image2:tag": "docker.url/image2:sometag",
        }
    }

    ImageImporter.imported_istags = []
    import_images(config_content, [])

    _check_oc_calls(mocker, mock_oc)


def test_images_mixed_style_syntax(mocker, mock_oc):
    config_content = {
        "images": [
            {"image1:tag": "docker.url/image1:sometag"},
            {"istag": "image2:tag", "from": "docker.url/image2:sometag"},
        ]
    }

    ImageImporter.imported_istags = []
    import_images(config_content, [])

    _check_oc_calls(mocker, mock_oc)


def test_images_conditional_images(mocker, mock_oc):
    config_content = {
        "images": [
            {"istag": "image1:tag", "from": "docker.url/image1:sometag", "envs": ["qa", "prod"]},
            {"istag": "image2:tag", "from": "docker.url/image2:sometag"},
        ]
    }

    ImageImporter.imported_istags = []
    import_images(config_content, ["prod"])

    _check_oc_calls(mocker, mock_oc)


def test_images_conditional_ignore_image(mocker, mock_oc):
    config_content = {
        "images": [
            {"istag": "image1:tag", "from": "docker.url/image1:sometag", "envs": ["qa", "prod"]},
            {"istag": "image2:tag", "from": "docker.url/image2:sometag"},
        ]
    }

    ImageImporter.imported_istags = []
    import_images(config_content, ["foo"])

    assert mock_oc.call_count == 1
    calls = [
        mocker.call(
            "import-image",
            "image2:tag",
            "--from=docker.url/image2:sometag",
            "--confirm",
            "--scheduled=True",
            _reraise=True,
        )
    ]
    mock_oc.assert_has_calls(calls)



import pytest

import ocdeployer.env


@pytest.fixture
def mock_files(monkeypatch, patch_os_path):
    def mock_get_cfg_files_in_dir(path):
        if path.endswith("empty_envTEST"):
            return []
        if path.endswith("envTEST"):
            return ["envTEST/test_envTEST.yml", "envTEST/other_envTEST.yml"]

    def mock_load_cfg_file(path):
        if path == "envTEST/test_envTEST.yml":
            return {
                "service": {
                    "enable_routes": False,
                    "enable_db": False,
                    "parameters": {"STUFF": "things"},
                }
            }
        if path == "envTEST/other_envTEST.yml":
            return {"another_service": {"somekey": "somevalue"}}

    # To understand why we're patching at 'ocdeployer.env'...
    # Read: https://docs.python.org/3/library/unittest.mock.html#where-to-patch
    monkeypatch.setattr("ocdeployer.env.get_cfg_files_in_dir", mock_get_cfg_files_in_dir)
    monkeypatch.setattr("ocdeployer.env.load_cfg_file", mock_load_cfg_file)


def test__load_vars_per_env(mock_files):
    handler = ocdeployer.env.EnvConfigHandler(
        env_names=["test_envTEST", "other_envTEST"], env_dir_name="envTEST"
    )

    expected = {
        "test_envTEST": {
            "service": {
                "enable_routes": False,
                "enable_db": False,
                "parameters": {"STUFF": "things"},
            }
        },
        "other_envTEST": {"another_service": {"somekey": "somevalue"}},
    }

    assert handler._load_vars_per_env() == expected


def test__load_vars_per_env_ignore_other_env(mock_files):
    handler = ocdeployer.env.EnvConfigHandler(env_names=["test_envTEST"], env_dir_name="envTEST")

    expected = {
        "test_envTEST": {
            "service": {
                "enable_routes": False,
                "enable_db": False,
                "parameters": {"STUFF": "things"},
            }
        }
    }

    assert handler._load_vars_per_env() == expected


def test__load_vars_per_env_no_envs_specified(mock_files):
    handler = ocdeployer.env.EnvConfigHandler(env_names=[], env_dir_name="envTEST")

    expected = {}

    assert handler._load_vars_per_env() == expected


def test__load_vars_per_env_no_env_files_in_dir(mock_files):
    handler = ocdeployer.env.EnvConfigHandler(
        env_names=["test_envTEST"], env_dir_name="empty_envTEST"
    )

    expected = {}

    assert handler._load_vars_per_env() == expected



import copy
import logging
import os
from collections import defaultdict

from cached_property import cached_property

from .config import merge_cfgs
from .utils import get_cfg_files_in_dir, get_dir, load_cfg_file, object_merge


log = logging.getLogger("ocdeployer.env")
GLOBAL = "global"
CFG = "_cfg"


def convert_to_regular_dict(data):
    if isinstance(data, defaultdict):
        data = {key: convert_to_regular_dict(val) for key, val in data.items()}
    return data


def nested_dict():
    return defaultdict(nested_dict)


def _dedupe_preserve_order(seq):
    """De-dupe a list, but preserve order of elements.

    https://www.peterbe.com/plog/uniqifiers-benchmark
    """
    seen = set()
    seen_add = seen.add
    return [x for x in seq if not (x in seen or seen_add(x))]


class EnvConfigHandler:
    def __init__(self, env_names, env_dir_name="env"):
        env_path = os.path.join(os.getcwd(), env_dir_name)
        self.base_env_path = get_dir(env_path, env_path, "environment")  # ensures path is valid dir
        self.env_dir_name = env_dir_name
        self.env_names = _dedupe_preserve_order(env_names)
        if len(env_names) != len(self.env_names):
            log.warning("Duplicate env names provided: %s", env_names)
        self._last_service_set = None
        self._last_merged_vars = None

    def _load_vars_per_env(self, path=None):
        data = {}

        if path:
            env_files = get_cfg_files_in_dir(path)
        else:
            env_files = get_cfg_files_in_dir(self.base_env_path)

        for file_path in env_files:
            env_name = os.path.splitext(os.path.basename(file_path))[0]
            if env_name not in self.env_names:
                continue

            data[env_name] = load_cfg_file(file_path)

        return data

    def _get_base_vars(self):
        """
        Load variables for env files located in the root 'env' dir

        Only files with a name listed in 'env_names' will be loaded.

        Returns a dict with keys/vals following this structure:
        {
            'env': {
                'service_set': {
                    'component': variables
                }
            }
        }

        "global" is a reserved service set name and component name
        """
        vars_per_env = self._load_vars_per_env()

        data = nested_dict()

        for env_name, env_vars in vars_per_env.items():
            for key, config in env_vars.items():
                if key == CFG:
                    # This is a env-level _cfg definition
                    data[env_name][CFG] = config
                elif key == GLOBAL:
                    # Global vars for all service sets
                    data[env_name][GLOBAL] = config
                elif "/" in key:
                    service_set = key.split("/")[0]
                    component = key.split("/")[1]
                    data[env_name][service_set][component] = config
                else:
                    # A specific component was not given, this is a service set var
                    # Global only for service set
                    service_set = key
                    data[env_name][service_set][GLOBAL] = config

        return convert_to_regular_dict(data)

    @cached_property
    def _base_vars(self):
        """
        Loads the base vars as a cached property, since they only need to be loaded once.
        """
        return self._get_base_vars()

    def _merge_environments(self, data):
        """
        Merge vars from multiple environments together

        The configs from the environments take precedence based on which order they were
        supplied to the command line. Environments listed first take precedence.

        Returns a dict with keys/vals following this structure:
        {
            'service_set': {
                'component': variables
            }
        }

        "global" is a reserved service set name and component name
        """
        merged_data = {}
        for env in self.env_names:
            object_merge(data.get(env, {}), merged_data)

        return merged_data

    def _get_service_set_vars(self, service_set_dir, service_set):
        """
        Load service set env data for each environment.
        """
        path = os.path.join(service_set_dir, self.env_dir_name)
        path = get_dir(path, path, "environment", optional="True")  # ensures path is valid dir

        vars_per_env = self._load_vars_per_env(path)

        data = nested_dict()

        for env_name, env_vars in vars_per_env.items():
            for component, variables in env_vars.items():
                if "/" in component:
                    # Service-set level env files should only be defining component sections, not
                    # "service_set/component" sections ... if we find a slash then strip out
                    # the leading service set name
                    component = component.split("/")[1]
                data[env_name][service_set][component] = variables

        return convert_to_regular_dict(data)

    def _merge_env_cfgs(self, vars_per_env, service_set=None):
        merged_cfg = {}
        for env in self.env_names:
            cfg = {}
            if service_set:
                # Look for a _cfg key under [env][service_set]['_cfg']
                for key, data in vars_per_env.get(env, {}).items():
                    if service_set and key == service_set:
                        cfg = data.get(CFG, {})
                        break
            else:
                # Look for a _cfg key under [env]['_cfg']
                cfg = vars_per_env.get(env, {}).get(CFG, {})
            merge_cfgs(cfg, merged_cfg)
        return merged_cfg

    def get_base_env_cfg(self):
        """
        Returns data defined under the '_cfg' key in the base env files.

        If _cfg is defined in multiple env files, its data is merged with precedence according to
        what order the envs were listed in.
        """
        return self._merge_env_cfgs(self._base_vars)

    def get_service_set_env_cfg(self, service_set_dir, service_set):
        """
        Returns data defined under the '_cfg' key in a service set's env files.

        If _cfg is defined in multiple env files, its data is merged with precedence according to
        what order the envs were listed in.
        """
        return self._merge_env_cfgs(
            self._get_service_set_vars(service_set_dir, service_set), service_set=service_set
        )

    def _merge_service_set_vars(self, service_set_dir, service_set):
        """
        Combine the env vars defined in a service set's env dir with the base env vars

        Returns a dict with keys/vals following this structure:
        {
            'env': {
                'service_set': {
                    'component': variables
                }
            }
        }

        "global" is a reserved service set name and component name
        """
        data = self._get_service_set_vars(service_set_dir, service_set)
        merged_vars = object_merge(copy.deepcopy(self._base_vars), data)
        self._last_service_set = service_set
        self._last_merged_vars = merged_vars

        # Don't include the '_cfg' component in this data set, it's not used for this purpose.
        if CFG in merged_vars:
            del merged_vars[CFG]

        return merged_vars

    def get_vars_for_component(self, service_set_dir, service_set, component):
        """
        Handles parsing of the variables data

        The base variables file is set up in the following way:

        global:
            VAR1: "blah"
            VAR2: "blah"

        advisor:
            VAR2: "this overrides global VAR2 for only components in the advisor set"

        advisor/advisor-db:
            VAR2: "this overrides global VAR2, and advisor VAR2, for only the advisor-db component"


        The service set variables file can be set up in the following way:
        global:
            VAR2: "this overrides VAR2 for all components in service set"

        advisor-db:
            VAR2: "this overrides VAR2 for only the advisor-db component in the service set"


        The base env vars file is merged with the service set level env vars file.

        Returns:
            dict of variables/values to apply to this specific component
        """
        if service_set == self._last_service_set:
            merged_vars = self._last_merged_vars
        else:
            merged_vars = self._merge_service_set_vars(service_set_dir, service_set)

        # Combine data from multiple env files (if provided) together
        merged_vars = self._merge_environments(merged_vars)

        component_level_vars = merged_vars.get(service_set, {}).get(component, {})
        service_set_level_vars = merged_vars.get(service_set, {}).get(GLOBAL, {})
        global_vars = merged_vars.get(GLOBAL, {})

        variables = copy.deepcopy(component_level_vars)
        if "parameters" not in variables:
            variables["parameters"] = {}

        variables = object_merge(service_set_level_vars, variables)
        variables = object_merge(global_vars, variables)

        return variables


class LegacyEnvConfigHandler(EnvConfigHandler):
    """
    Allows use of --env in "legacy mode", i.e. pass in specific env files instead of env names.
    """

    @staticmethod
    def _get_env_name(file_path):
        return os.path.splitext(os.path.basename(file_path))[0]

    def __init__(self, env_files, env_dir_name="env"):
        self.env_files = env_files
        _env_names = [self._get_env_name(fp) for fp in self.env_files]
        super().__init__(_env_names, env_dir_name)

    def _load_vars_per_env(self, path=None):
        data = {}

        for file_path in self.env_files:
            data[self._get_env_name(file_path)] = load_cfg_file(file_path)

        return data

    def _merge_service_set_vars(self, env_dir_path, service_set):
        return self._base_vars



"""
Contains utils such as update checker

"""
from distutils.version import StrictVersion

import pkg_resources
import requests

PYPI_URL = "https://pypi.python.org/pypi/{}/json"


def _compare_version(pypi_version, pkg_name):
    pypi_version = StrictVersion(pypi_version)
    try:
        local_version = pkg_resources.get_distribution(pkg_name).version
    except pkg_resources.DistributionNotFound:
        local_version = "0.0.0"

    try:
        my_version = StrictVersion(local_version)
    except ValueError:
        print(f"Version '{local_version}' seems to be a dev version, assuming up-to-date")
        return

    if my_version < pypi_version:
        print(
            f"\nThere is a new version available! (yours: {my_version}, available: {pypi_version})"
            f"\nTo upgrade:\n\t$ pip install --upgrade {pkg_name}"
        )
    else:
        print(f"Up-to-date! (version {my_version})")


def check_pypi(pkg_name, timeout=5):
    print(f"\nChecking pypi for latest release of {pkg_name}...")

    pkg_data = {}
    try:
        response = requests.get(PYPI_URL.format(pkg_name), timeout=timeout)
        response.raise_for_status()
        pkg_data = response.json()
    except requests.exceptions.Timeout:
        print("Unable to reach pypi quickly, giving up.")
    except requests.exceptions.HTTPError as e:
        print("Error response from pypi: ", e.errno, e.message)
    except ValueError:
        print("Response was not valid json, giving up.")

    try:
        pypi_version = pkg_data["info"]["version"]
    except KeyError:
        print("Unable to parse version info from pypi")
    else:
        _compare_version(pypi_version, pkg_name)
    print("\n")



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="ocdeployer",
    use_scm_version=True,
    description=(
        "A tool which wraps the OpenShift command line tools to enable repeatable automated"
        " deployment of OpenShift templates"
    ),
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    author="Brandon Squizzato",
    author_email="bsquizza@redhat.com",
    url="https://www.github.com/bsquizz/ocdeployer",
    packages=find_packages(),
    keywords=["openshift", "kubernetes"],
    setup_requires=["setuptools_scm"],
    include_package_data=True,
    install_requires=[
        "sh",
        "prompter",
        "pyyaml",
        "click",
        "appdirs",
        "wait_for",
        "jinja2",
        "cached_property",
        "pytz",
        "kubernetes",
        "anytree",
    ],
    classifiers=[
        "Topic :: Utilities",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
    ],
    python_requires=">=3.5",
)



import pytest
import os

from ocdeployer.secrets import SecretImporter
from ocdeployer.deploy import DeployRunner
from ocdeployer.env import EnvConfigHandler, LegacyEnvConfigHandler


def patched_runner(env_values, mock_load_vars_per_env, legacy=False):
    if not env_values:
        handler = None
    elif legacy:
        handler = LegacyEnvConfigHandler(env_files=env_values, env_dir_name="envTEST")
        handler.env_names = env_values
    else:
        handler = EnvConfigHandler(env_names=env_values, env_dir_name="envTEST")

    runner = DeployRunner(
        "templatesTEST", "test-project", handler, None, ["service"], None, None, []
    )
    runner.base_env_path = "base/envTEST"

    if handler:
        runner.env_config_handler._load_vars_per_env = mock_load_vars_per_env
    return runner


def build_mock_env_loader(base_env_data, service_set_env_data={}):
    def mock_load_vars_per_env(path=None):
        print(f"Mock loader received path: {path}")
        if path is None:
            return base_env_data
        if "base" in path and path.endswith("envTEST"):
            print("Loading mock base data")
            return base_env_data
        if "templates" in path and "service" in path and path.endswith("envTEST"):
            print("Loading mock service set data")
            return service_set_env_data
        return {}

    return mock_load_vars_per_env


@pytest.fixture
def patch_load_cfg(monkeypatch):
    """
    A fixture that returns a function which will patch 'utils.load_cfg_file' when called in a test

    The caller can specify the dict that should be returned when 'load_cfg_file' is called
    against these two paths:
    * "templatesTEST/_cfg.yml"
    * "templatesTEST/service/_cfg.yml"
    """

    def _func(base_cfg_data, service_cfg_data):
        def _patched_load_cfg_file(path):
            if path.endswith(os.path.join("templatesTEST", "service", "_cfg.yml")):
                return service_cfg_data
            if path.endswith(os.path.join("templatesTEST", "_cfg.yml")):
                return base_cfg_data
            else:
                raise Exception("Unknown path passed to load_cfg_file")

        monkeypatch.setattr("ocdeployer.deploy.load_cfg_file", _patched_load_cfg_file)

    yield _func


def test_cfg_no_env_given(patch_os_path, patch_load_cfg):
    runner = patched_runner(None, None)
    base_cfg_data = {
        "secrets": ["secret1"],
        "images": [{"image1:latest": "somerepo/image1:latest"}],
    }
    set_cfg_data = {"secrets": ["secret2"], "images": [{"image2:latest": "somerepo/image2:latest"}]}
    patch_load_cfg(base_cfg_data, set_cfg_data)

    # no env file defined that has '_cfg' key, there should be no changes to the cfg
    assert runner._get_base_cfg() == {
        "secrets": [{"name": "secret1", "envs": [], "link": []}],
        "images": [
            {
                "istag": "image1:latest",
                "from": "somerepo/image1:latest",
                "envs": [],
                "scheduled": True,
            }
        ],
    }

    assert runner._get_service_set_cfg("service", "templatesTEST/service") == {
        "secrets": [{"name": "secret2", "envs": [], "link": []}],
        "images": [
            {
                "istag": "image2:latest",
                "from": "somerepo/image2:latest",
                "envs": [],
                "scheduled": True,
            }
        ],
    }


def test_cfg_no_env_cfg(patch_os_path, patch_load_cfg):
    base_cfg_data = {
        "secrets": ["secret1"],
        "images": [{"image1:latest": "somerepo/image1:latest"}],
    }
    set_cfg_data = {"secrets": ["secret2"], "images": [{"image2:latest": "somerepo/image2:latest"}]}
    mock_var_data = {"test_env": {"service": {"some_set": "stuff"}}}
    mock_set_var_data = {"test_env": {"component": {"some_var": "stuff"}}}

    patch_load_cfg(base_cfg_data, set_cfg_data)
    runner = patched_runner(["test_env"], build_mock_env_loader(mock_var_data, mock_set_var_data))

    # env file had no _cfg, so we should not see any changes to the cfg
    assert runner._get_base_cfg() == {
        "secrets": [{"name": "secret1", "envs": [], "link": []}],
        "images": [
            {
                "istag": "image1:latest",
                "from": "somerepo/image1:latest",
                "envs": [],
                "scheduled": True,
            }
        ],
    }

    assert runner._get_service_set_cfg("service", "templatesTEST/service") == {
        "secrets": [{"name": "secret2", "envs": [], "link": []}],
        "images": [
            {
                "istag": "image2:latest",
                "from": "somerepo/image2:latest",
                "envs": [],
                "scheduled": True,
            }
        ],
    }


def test_cfg_base_env_cfg(patch_os_path, patch_load_cfg):
    base_cfg_data = {
        "secrets": ["secret1"],
        "images": [{"image1:latest": "somerepo/image1:latest"}],
    }
    set_cfg_data = {"secrets": ["secret2"], "images": [{"image2:latest": "somerepo/image2:latest"}]}
    mock_var_data = {
        "test_env": {
            "_cfg": {
                "secrets": ["additional-secret1"],
                "images": [{"image1:latest": "overridden-image"}],
                "extrastuff": "things",
            },
            "service": {"some_set": "stuff"},
        }
    }
    mock_set_var_data = {"test_env": {"component": {"some_var": "stuff"}}}

    patch_load_cfg(base_cfg_data, set_cfg_data)
    runner = patched_runner(["test_env"], build_mock_env_loader(mock_var_data, mock_set_var_data))

    # base env file had a _cfg, we should see base env file config merged into base config
    assert runner._get_base_cfg() == {
        "secrets": [
            {"name": "additional-secret1", "envs": [], "link": []},
            {"name": "secret1", "envs": [], "link": []},
        ],
        "extrastuff": "things",
        "images": [
            {"istag": "image1:latest", "from": "overridden-image", "envs": [], "scheduled": True}
        ],
    }

    # service set env file had no _cfg, so it should be unchanged
    assert runner._get_service_set_cfg("service", "templatesTEST/service") == {
        "secrets": [{"name": "secret2", "envs": [], "link": []}],
        "images": [
            {
                "istag": "image2:latest",
                "from": "somerepo/image2:latest",
                "envs": [],
                "scheduled": True,
            }
        ],
    }


def test_cfg_set_env_cfg(patch_os_path, patch_load_cfg):
    base_cfg_data = {
        "secrets": ["secret1"],
        "images": [{"image1:latest": "somerepo/image1:latest"}],
    }
    set_cfg_data = {"secrets": ["secret2"], "images": [{"image2:latest": "somerepo/image2:latest"}]}
    mock_var_data = {
        "test_env": {
            "_cfg": {
                "secrets": ["additional-secret1"],
                "images": [{"image1:latest": "overridden-image"}],
                "extrastuff": "things",
            },
            "service": {"some_set": "stuff"},
        }
    }
    mock_set_var_data = {
        "test_env": {
            "_cfg": {
                "secrets": ["some-secret-for-set-only"],
                "images": [
                    {"image3": "somerepo/image3:latest"},
                    {"image2:latest": "overridden-image-2"},
                ],
            },
            "component": {"some_var": "stuff"},
        }
    }

    patch_load_cfg(base_cfg_data, set_cfg_data)
    runner = patched_runner(["test_env"], build_mock_env_loader(mock_var_data, mock_set_var_data))

    # base env file had a _cfg, we should see base env file config merged into base config
    assert runner._get_base_cfg() == {
        "secrets": [
            {"name": "additional-secret1", "envs": [], "link": []},
            {"name": "secret1", "envs": [], "link": []},
        ],
        "extrastuff": "things",
        "images": [
            {"istag": "image1:latest", "from": "overridden-image", "envs": [], "scheduled": True}
        ],
    }

    # set env file had a _cfg, we should see set env file config merged into set config
    assert runner._get_service_set_cfg("service", "templatesTEST/service") == {
        "secrets": [
            {"name": "some-secret-for-set-only", "envs": [], "link": []},
            {"name": "secret2", "envs": [], "link": []},
        ],
        "images": [
            {"istag": "image2:latest", "from": "overridden-image-2", "envs": [], "scheduled": True},
            {
                "istag": "image3:latest",
                "from": "somerepo/image3:latest",
                "envs": [],
                "scheduled": True,
            },
        ],
    }


def test__no_env_given():
    expected = {
        "parameters": {
            "NAMESPACE": "test-project",
            "SECRETS_PROJECT": SecretImporter.source_project,
        },
    }

    runner = patched_runner(None, None, legacy=False)
    assert runner._get_variables("service", "templates/service", "some_component") == expected


@pytest.mark.parametrize("legacy", (True, False), ids=("legacy=true", "legacy=false"))
def test__get_variables_sanity(legacy, patch_os_path):
    mock_var_data = {
        "test_env": {
            "service": {
                "enable_routes": False,
                "enable_db": False,
                "parameters": {"STUFF": "things"},
            }
        }
    }

    expected = {
        "enable_routes": False,
        "enable_db": False,
        "parameters": {
            "STUFF": "things",
            "NAMESPACE": "test-project",
            "SECRETS_PROJECT": SecretImporter.source_project,
        },
    }

    runner = patched_runner(["test_env"], build_mock_env_loader(mock_var_data), legacy)
    assert runner._get_variables("service", "templates/service", "some_component") == expected


@pytest.mark.parametrize("legacy", (True, False), ids=("legacy=true", "legacy=false"))
def test__get_variables_merge_from_global(legacy, patch_os_path):
    mock_var_data = {
        "test_env": {
            "global": {"global_variable": "global-value", "parameters": {"GLOBAL": "things"}},
            "service": {"service_variable": True, "parameters": {"STUFF": "service-stuff"}},
            "service/component": {
                "component_variable": "component",
                "parameters": {"COMPONENT": "component-param"},
            },
        }
    }

    expected = {
        "component_variable": "component",
        "global_variable": "global-value",
        "service_variable": True,
        "parameters": {
            "COMPONENT": "component-param",
            "GLOBAL": "things",
            "STUFF": "service-stuff",
            "NAMESPACE": "test-project",
            "SECRETS_PROJECT": SecretImporter.source_project,
        },
    }

    runner = patched_runner(["test_env"], build_mock_env_loader(mock_var_data), legacy)
    assert runner._get_variables("service", "templates/service", "component") == expected


@pytest.mark.parametrize("legacy", (True, False), ids=("legacy=true", "legacy=false"))
def test__get_variables_service_overwrite_parameter(legacy, patch_os_path):
    mock_var_data = {
        "test_env": {
            "global": {"parameters": {"STUFF": "things"}},
            "service": {"parameters": {"STUFF": "service-stuff"}},
        }
    }

    expected = {
        "parameters": {
            "STUFF": "service-stuff",
            "NAMESPACE": "test-project",
            "SECRETS_PROJECT": SecretImporter.source_project,
        }
    }

    runner = patched_runner(["test_env"], build_mock_env_loader(mock_var_data), legacy)
    assert runner._get_variables("service", "templates/service", "component") == expected


@pytest.mark.parametrize("legacy", (True, False), ids=("legacy=true", "legacy=false"))
def test__get_variables_service_overwrite_variable(legacy, patch_os_path):
    mock_var_data = {"test_env": {"global": {"enable_db": False}, "service": {"enable_db": True}}}

    expected = {
        "enable_db": True,
        "parameters": {
            "NAMESPACE": "test-project",
            "SECRETS_PROJECT": SecretImporter.source_project,
        },
    }

    runner = patched_runner(["test_env"], build_mock_env_loader(mock_var_data), legacy)
    assert runner._get_variables("service", "templates/service", "component") == expected


@pytest.mark.parametrize("legacy", (True, False), ids=("legacy=true", "legacy=false"))
def test__get_variables_component_overwrite_parameter(legacy, patch_os_path):
    mock_var_data = {
        "test_env": {
            "global": {"parameters": {"STUFF": "things"}},
            "service": {"parameters": {"THINGS": "service-things"}},
            "service/component": {"parameters": {"THINGS": "component-things"}},
        }
    }

    expected = {
        "parameters": {
            "STUFF": "things",
            "THINGS": "component-things",
            "NAMESPACE": "test-project",
            "SECRETS_PROJECT": SecretImporter.source_project,
        }
    }

    runner = patched_runner(["test_env"], build_mock_env_loader(mock_var_data), legacy)
    assert runner._get_variables("service", "templates/service", "component") == expected


@pytest.mark.parametrize("legacy", (True, False), ids=("legacy=true", "legacy=false"))
def test__get_variables_component_overwrite_variable(legacy, patch_os_path):
    mock_var_data = {
        "test_env": {
            "global": {"enable_routes": False},
            "service": {"enable_db": True},
            "service/component": {"enable_db": False},
        }
    }

    expected = {
        "enable_routes": False,
        "enable_db": False,
        "parameters": {
            "NAMESPACE": "test-project",
            "SECRETS_PROJECT": SecretImporter.source_project,
        },
    }

    runner = patched_runner(["test_env"], build_mock_env_loader(mock_var_data), legacy)
    assert runner._get_variables("service", "templates/service", "component") == expected


def test__get_variables_base_and_service_set(patch_os_path):
    base_var_data = {
        "test_env": {
            "global": {"global_var": "base_global", "parameters": {"GLOBAL_PARAM": "things"}}
        }
    }

    service_set_var_data = {
        "test_env": {
            "global": {"global_set_var": "set_global", "parameters": {"PARAM": "something"}},
            "component": {"component_var": "something", "parameters": {"ANOTHER_PARAM": "stuff"}},
        }
    }

    expected = {
        "global_var": "base_global",
        "global_set_var": "set_global",
        "component_var": "something",
        "parameters": {
            "GLOBAL_PARAM": "things",
            "PARAM": "something",
            "ANOTHER_PARAM": "stuff",
            "NAMESPACE": "test-project",
            "SECRETS_PROJECT": SecretImporter.source_project,
        },
    }

    runner = patched_runner(
        ["test_env"], build_mock_env_loader(base_var_data, service_set_var_data)
    )
    assert runner._get_variables("service", "templates/service", "component") == expected


def test__get_variables_service_set_only(patch_os_path):
    base_var_data = {}

    service_set_var_data = {
        "test_env": {
            "global": {"global_set_var": "set_global", "parameters": {"PARAM": "something"}},
            "component": {"component_var": "something", "parameters": {"ANOTHER_PARAM": "stuff"}},
        }
    }

    expected = {
        "global_set_var": "set_global",
        "component_var": "something",
        "parameters": {
            "PARAM": "something",
            "ANOTHER_PARAM": "stuff",
            "NAMESPACE": "test-project",
            "SECRETS_PROJECT": SecretImporter.source_project,
        },
    }

    runner = patched_runner(
        ["test_env"], build_mock_env_loader(base_var_data, service_set_var_data)
    )
    assert runner._get_variables("service", "templates/service", "component") == expected


def test__get_variables_service_set_overrides(patch_os_path):
    base_var_data = {
        "test_env": {
            "global": {"global_var": "base_global", "parameters": {"GLOBAL_PARAM": "things"}},
            "service": {"global_set_var": "blah", "parameters": {"PARAM": "blah"}},
            "service/component": {"component_var": "override this"},
        }
    }

    service_set_var_data = {
        "test_env": {
            "global": {"global_set_var": "set_global", "parameters": {"PARAM": "something"}},
            "component": {"component_var": "something", "parameters": {"ANOTHER_PARAM": "stuff"}},
        }
    }

    expected = {
        "global_var": "base_global",
        "global_set_var": "set_global",
        "component_var": "something",
        "parameters": {
            "GLOBAL_PARAM": "things",
            "PARAM": "something",
            "ANOTHER_PARAM": "stuff",
            "NAMESPACE": "test-project",
            "SECRETS_PROJECT": SecretImporter.source_project,
        },
    }

    runner = patched_runner(
        ["test_env"], build_mock_env_loader(base_var_data, service_set_var_data)
    )
    assert runner._get_variables("service", "templates/service", "component") == expected


def test__get_variables_multiple_envs(patch_os_path):
    base_var_data = {
        "test_env": {
            "global": {"global_var": "base_global1", "parameters": {"GLOBAL_PARAM": "things1"}},
        },
        "test_env2": {
            "global": {"global_var": "base_global2"},
            "service/component": {"component_var": "comp2"},
        },
        "test_env3": {
            "global": {
                "global_var": "base_global3",
                "parameters": {"GLOBAL_PARAM": "things3", "ENV3_PARAM": "env3"},
            },
            "service/component": {"component_var": "comp3"},
        },
    }

    service_set_var_data = {
        "test_env": {"global": {"global_set_var": "set_global1"}},
        "test_env2": {
            "service/component": {
                "component_var": "comp2-set",
                "parameters": {"ENV2_PARAM": "env2"},
            }
        },
    }

    expected = {
        "global_var": "base_global1",
        "global_set_var": "set_global1",
        "component_var": "comp2-set",
        "parameters": {
            "GLOBAL_PARAM": "things1",
            "ENV3_PARAM": "env3",
            "ENV2_PARAM": "env2",
            "NAMESPACE": "test-project",
            "SECRETS_PROJECT": SecretImporter.source_project,
        },
    }

    runner = patched_runner(
        ["test_env", "test_env2", "test_env3"],
        build_mock_env_loader(base_var_data, service_set_var_data),
    )
    assert runner._get_variables("service", "templates/service", "component") == expected


def test__get_variables_multiple_envs_legacy(patch_os_path):
    base_var_data = {
        "test_env": {
            "global": {"global_var": "base_global1", "parameters": {"GLOBAL_PARAM": "things1"}},
        },
        "test_env2": {
            "global": {"global_var": "base_global2"},
            "service/component": {"component_var": "comp2"},
        },
        "test_env3": {
            "global": {
                "global_var": "base_global3",
                "parameters": {"GLOBAL_PARAM": "things3", "ENV3_PARAM": "env3"},
            },
            "service/component": {"component_var": "comp3"},
        },
    }

    expected = {
        "global_var": "base_global1",
        "component_var": "comp2",
        "parameters": {
            "GLOBAL_PARAM": "things1",
            "ENV3_PARAM": "env3",
            "NAMESPACE": "test-project",
            "SECRETS_PROJECT": SecretImporter.source_project,
        },
    }

    runner = patched_runner(
        ["test_env", "test_env2", "test_env3"], build_mock_env_loader(base_var_data), legacy=True
    )
    assert runner._get_variables("service", "templates/service", "component") == expected


def test__get_variables_multiple_envs_precedence(patch_os_path):
    base_var_data = {"test_env1": {"service/component": {"parameters": {"PARAM": "things1"}}}}

    service_set_var_data = {"test_env2": {"component": {"parameters": {"PARAM": "things2"}}}}

    expected = {
        "parameters": {
            "PARAM": "things1",
            "NAMESPACE": "test-project",
            "SECRETS_PROJECT": SecretImporter.source_project,
        },
    }

    runner = patched_runner(
        ["test_env1", "test_env2"], build_mock_env_loader(base_var_data, service_set_var_data),
    )
    assert runner._get_variables("service", "templates/service", "component") == expected


def test__get_variables_multiple_envs_precedence_reversed(patch_os_path):
    base_var_data = {"test_env1": {"service/component": {"parameters": {"PARAM": "things1"}}}}

    service_set_var_data = {"test_env2": {"component": {"parameters": {"PARAM": "things2"}}}}

    expected = {
        "parameters": {
            "PARAM": "things2",
            "NAMESPACE": "test-project",
            "SECRETS_PROJECT": SecretImporter.source_project,
        },
    }

    runner = patched_runner(
        ["test_env2", "test_env1"], build_mock_env_loader(base_var_data, service_set_var_data),
    )
    assert runner._get_variables("service", "templates/service", "component") == expected
